
import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import pytest
import pandas as pd
import dreem.util as util
import os
from dreem import clustering
import json 
import numpy as np

module = 'clustering'

sample_name = 'test_set_1'
MAX_CLUSTERS = 2

reads_partition = [100000]*2

half_sequence_length = 100
unpaired_bases = int(0.3*half_sequence_length)
shared_bases = [int(u*unpaired_bases) for u in np.linspace(0.0, 0.8, 4)]
mu_unpaired = [0.07]

# sample_profile = {}
# for sc in shared_bases:
#     for mu in mu_unpaired:
#         profile_name = 'r{}_ub{}_sb{}_mu{}'.format(reads_partition, unpaired_bases, sc, int(mu*100))
#         sample_profile[profile_name] = {
#             'n_reads': reads_partition,
#             'n_AC': half_sequence_length,
#             'n_unpaired': unpaired_bases,
#             'n_shared': sc,
#             'path_bv': os.path.join(test_files_dir, 'input', module, sample_name, 'r{}_ub{}_sb{}_mu{}/0.orc'.format(reads_partition, unpaired_bases, sc, int(mu*100))),
#             'path_json': os.path.join(test_files_dir, 'output', module, sample_name, profile_name),
#             'mu_unpaired': [mu]*2
#         }
    
# module_input = os.path.join(input_dir, module)
# module_expected = os.path.join(prediction_dir, module)
# module_output =  os.path.join(output_dir, module)

# inputs = ['clustering']

# @pytest.mark.skip(reason="Too bugged")
# def test_make_files():
#     os.makedirs(os.path.join(test_files_dir, 'input', module, sample_name), exist_ok=True)
#     os.makedirs(os.path.join(test_files_dir, 'expected_output', module, sample_name), exist_ok=True)
#     files_generator_library.generate_files(sample_profile, module, inputs, [], test_files_dir, sample_name)
#     files_generator_library.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    
# @pytest.mark.skip(reason="Too bugged")
# def test_run():
#     clustering.run(
#         input_dir = os.path.join(module_input, sample_name),
#         out_dir = module_output,
#         num_runs= 5,
#         min_iter=30,
#         max_clusters=MAX_CLUSTERS,
#         n_cpus=10
#     )

# @pytest.mark.skip(reason="Too bugged")
# def test_assert_files_exist():
#     assert os.path.exists(os.path.join(module_output, 'best_cluster_reads.json'))

# @pytest.mark.skip(reason="not implemented")
# def test_assess_performance():

#     for name, sample in sample_profile.items():
    
#         with open(os.path.join(module_output, 'best_cluster_reads.json'), 'r') as f:
#             result_dict = json.load(f)
#             n_reads = [ sum([k[1]==str(idx+1) for k in result_dict[name]['K2'].keys()]) for idx in range(len(sample['n_reads']))]

#             result_np = pd.DataFrame.from_dict(result_dict[name]['K'+str(len(sample['n_reads']))]).to_numpy().T
#             F1 = 1.0

#             if len(n_reads) == 2:
#                 F1_class = []
#                 for i in range(2):
#                     TP = np.count_nonzero(np.argmax(result_np[:n_reads[0]], axis=1)==i)
#                     if TP != 0:
#                         precision = TP/np.count_nonzero(np.argmax(result_np, axis=1)==i)
#                         recall = TP/n_reads[0]
#                         F1_class.append(2*precision*recall/(precision+recall))
#                 F1 = max(F1_class)
    
#             if len(n_reads) == 3:
#                 n_reads_cumulative = [0] + n_reads[:-1]
#                 F1 = 0
#                 indices = [0,1,2]
                
#                 # Compute F1 score of each class for all combinations
#                 for c in range(len(n_reads)):
#                     F1_class = []
#                     start = sum(n_reads_cumulative[:c+1])
#                     end = start + n_reads[c]
#                     for i in indices:
#                         TP = np.count_nonzero(np.argmax(result_np[start:end], axis=1)==i)
#                         if TP != 0:
#                             precision = TP/np.count_nonzero(np.argmax(result_np, axis=1)==i)
#                             recall = TP/n_reads[c]
#                             F1_class.append(2*precision*recall/(precision+recall))
#                         else:
#                             F1_class.append(0.0)
#                     indices.pop(np.argmax(F1_class))
#                     F1 += np.max(F1_class)
#                 F1 /= 3

#             print('F1 score: ', F1)
#             sample['F1'] = F1
#             assert F1 > 0.5

                

# if __name__ == '__main__':
#     test_make_files()
#     test_run()
#     test_assert_files_exist()
#     test_assess_performance()



from __future__ import annotations
from functools import cached_property, wraps
from io import BufferedReader
import logging
from typing import Callable, Optional

from ..align.reads import BamVectorSelector, SamVectorSorter, SAM_HEADER
from ..util.path import OneRefAlignmentInFilePath, OneRefAlignmentStepFilePath
from ..vector.vector import *


def _reset_seek(func: Callable):
    @wraps(func)
    def wrapper(reader: SamReader, *args, **kwargs):
        prev_pos = reader.sam_file.tell()
        result = func(reader, *args, **kwargs)
        reader.sam_file.seek(prev_pos)
        return result
    return wrapper


def _range_of_records(get_records_func: Callable):
    @wraps(get_records_func)
    @_reset_seek
    def wrapper(reader: SamReader, start: int, stop: int | None):
        reader.sam_file.seek(start)
        records = get_records_func(reader)
        if stop is None:
            yield from records
        else:
            while True:
                if reader.sam_file.tell() < stop:
                    yield next(records)
                else:
                    assert reader.sam_file.tell() == stop
                    break
    return wrapper


@_range_of_records
def _get_records_single(reader: SamReader):
    while read := SamRead(reader.sam_file.readline()):
        if read.flag.paired:
            logging.error(f"Got paired-end read {read} in single-end SAM file")
        else:
            yield SamRecord(read)


@_range_of_records
def _get_records_paired_lenient(reader: SamReader):
    prev_read: Optional[SamRead] = None
    while line := reader.sam_file.readline():
        read = SamRead(line)
        if not read.flag.paired:
            logging.error(f"Got single-end read {read} in paired-end SAM file")
            continue
        if prev_read:
            # The previous read has not yet been yielded
            if prev_read.qname == read.qname:
                # The current read is the mate of the previous read
                if prev_read.flag.first:
                    yield SamRecord(prev_read, read)
                else:
                    yield SamRecord(read, prev_read)
                prev_read = None
            else:
                # The previous read is paired, but its mate is not
                # in the SAM file. This situation can happen when
                # only the reads mapping to a specific region are
                # exported to a SAM file and only one mate overlaps
                # that region. In this case, it is valid for only
                # one of the two paired mates to be given. Thus,
                # strict mode (which normally ensures that read 2
                # is given if read 1 is paired) is turned off. The
                # given mate is processed as if a single-end read.
                yield SamRecord(prev_read, strict=False)
                # Save the current read so that if its mate is the next
                # read, it will be returned as a pair.
                prev_read = read
        else:
            # Save the current read so that if its mate is the next
            # read, it will be returned as a pair.
            prev_read = read
    if prev_read:
        # In case the last record
        yield SamRecord(prev_read)


@_range_of_records
def _get_records_paired_strict(reader: SamReader):
    while line := reader.sam_file.readline():
        yield SamRecord(SamRead(line),
                        SamRead(reader.sam_file.readline()),
                        strict=True)


class SamReader(object):
    """ Interface to a SAM file. Extract reads from a specific region
    of a BAM file into a SAM file, then return a SamRecord object for
    each record in the file. """

    def __init__(self,
                 xam_path: OneRefAlignmentInFilePath,
                 temp_dir: str,
                 end5: int,
                 end3: int,
                 spans: bool,
                 n_procs: int,
                 save_temp: bool,
                 resume: bool,
                 owner: bool):
        self.xam_path = xam_path
        self.temp_dir = temp_dir
        self.end5 = end5
        self.end3 = end3
        self.spans = spans
        self.n_procs = n_procs
        self.save_temp = save_temp
        self.resume = resume
        self.owner = owner
        self._sam_sorter: SamVectorSorter | None = None
        self._sam_path: OneRefAlignmentStepFilePath | None = None
        self._sam_file: BufferedReader | None = None
    
    def __enter__(self):
        # Convert the BAM file to a temporary SAM file
        if self.owner:
            selector = BamVectorSelector(top_dir=self.temp_dir,
                                         save_temp=self.save_temp,
                                         resume=self.resume,
                                         n_procs=self.n_procs,
                                         input_path=self.xam_path,
                                         ref=self.ref,
                                         end5=self.end5,
                                         end3=self.end3)
            if selector.output.path.is_file() and self.resume:
                xam_selected = selector.output
            else:
                xam_selected = selector.run()
            sorter = SamVectorSorter(top_dir=self.temp_dir,
                                     save_temp=self.save_temp,
                                     resume=self.resume,
                                     n_procs=self.n_procs,
                                     input_path=xam_selected)
            if sorter.output.path.is_file() and self.resume:
                self._sam_path = sorter.output
            else:
                self._sam_path = sorter.run(name=True)
            if not self.save_temp:
                selector.clean()
        else:
            self._sam_path = self.xam_path
        self._sam_file = open(self.sam_path.path, "rb")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._sam_file.close()
        self._sam_file = None
        if self._sam_sorter is not None and not self.save_temp:
            self._sam_sorter.clean()
        if self.owner and not self.save_temp:
            self.sam_path.path.unlink(missing_ok=True)
            self._sam_path = None

    @property
    def sample(self):
        return self.xam_path.sample

    @property
    def ref(self):
        return self.xam_path.ref

    @property
    def sam_path(self):
        if not self._sam_path:
            raise RuntimeError(f"{self} has no open SAM path.")
        return self._sam_path

    @property
    def sam_file(self):
        if not self._sam_file:
            raise RuntimeError(f"{self} has no open SAM file.")
        return self._sam_file

    def _seek_beginning(self):
        self.sam_file.seek(0)
    
    @cached_property
    @_reset_seek
    def _rec1_pos(self):
        self._seek_beginning()
        while (line := self.sam_file.readline()).startswith(SAM_HEADER):
            pass
        return self.sam_file.tell() - len(line)

    def _seek_rec1(self):
        self.sam_file.seek(self._rec1_pos)

    @cached_property
    @_reset_seek
    def paired(self):
        self._seek_rec1()
        first_line = self.sam_file.readline()
        return SamRead(first_line).flag.paired if first_line else False
    
    def get_records(self, start: int, stop: int, strict_pairs: bool):
        # Each record_generator method is obtained by self.__class__
        # instead of just self because the method is static
        if self.paired:
            if strict_pairs:
                if self.spans:
                    return _get_records_paired_strict(self, start, stop)
                logging.warning(f"Disabling strict pairs for {self} because "
                                f"it does not span the reference sequence")
            return _get_records_paired_lenient(self, start, stop)
        return _get_records_single(self, start, stop)
    
    @_reset_seek
    def get_batch_indexes(self, vectors_per_batch: int):
        if vectors_per_batch <= 0:
            raise ValueError("vectors_per_batch must be a positive integer")
        n_skip = (self.paired + 1) * (vectors_per_batch - 1)
        self._seek_rec1()
        while True:
            yield self.sam_file.tell()
            try:
                line = next(self.sam_file)
            except StopIteration:
                break
            else:
                for _, line in zip(range(n_skip), self.sam_file):
                    pass
                if self.paired:
                    # Compare the current and next query names.
                    try:
                        qname = line.split()[0]
                    except IndexError:
                        pass
                    else:
                        line_next = self.sam_file.readline()
                        try:
                            qname_next = line_next.split()[0]
                        except IndexError:
                            pass
                        else:
                            if qname != qname_next:
                                # If the current and next query names differ
                                # (the lines do not come from two paired mates),
                                # then backtrack to the beginning of line_next.
                                self.sam_file.seek(-len(line_next), 1)

    def __str__(self):
        return f"{self.sample}@{self.ref}:{self.end5}-{self.end3}"



from collections import defaultdict
from functools import partial
import itertools
import logging
from multiprocessing import Pool

from ..align.reads import (BamAlignSorter, BamOutputter, BamSplitter,
                           FastqAligner, FastqTrimmer, FastqUnit,
                           SamRemoveEqualMappers)
from ..util import path
from ..util import docdef
from ..util.seq import FastaParser, FastaWriter
from ..util.util import get_num_parallel


def check_for_duplicates(fq_units: list[FastqUnit]):
    # Count the number of times each sample and reference occurs.
    samples = defaultdict(int)
    sample_refs = defaultdict(lambda: defaultdict(int))
    for fq_unit in fq_units:
        if fq_unit.by_ref:
            sample_refs[fq_unit.sample][fq_unit.ref] += 1
        else:
            samples[fq_unit.sample] += 1
    # Find duplicates.
    dups = set()
    # Duplicate whole-sample FASTQs
    dups = dups | {sample for sample, count in samples.items() if count > 1}
    # Duplicate demultiplexed FASTQs
    dups = dups | {(sample, ref) for sample, refs in sample_refs.items()
                   for ref, count in refs.items() if count > 1}
    # Duplicate samples between whole-sample and demultiplexed FASTQs
    dups = dups | (set(samples) & set(sample_refs))
    return dups


def write_temp_ref_files(temp_dir: str,
                         fasta: str,
                         fq_units: list[FastqUnit]):
    """ Write temporary FASTA files, each containing one reference that
    corresponds to a FASTQ file from demultiplexing. """
    ref_paths: dict[str, path.OneRefSeqStepFilePath] = dict()
    # Only the reference sequences of FASTQ files that have come from
    # demultiplexing need to be written.
    refs = {fq_unit.ref for fq_unit in fq_units if fq_unit.by_ref}
    if refs:
        # Parse the FASTA only if there are any references to write.
        for ref, seq in FastaParser(fasta).parse():
            if ref in refs:
                # Write the reference sequence to a temporary FASTA file
                # only if at least one demultiplexed FASTQ file uses it.
                ref_path = path.OneRefSeqStepFilePath(
                    top=temp_dir,
                    module=path.Module.ALIGN.value,
                    step=path.Step.ALIGN_ALIGN.value,
                    ref=ref,
                    ext=path.FASTA_EXTS[0])
                ref_path.path.parent.mkdir(parents=True, exist_ok=True)
                logging.info(f"Writing temporary FASTA file: {ref_path}")
                FastaWriter(ref_path.path, {ref: seq}).write()
                ref_paths[ref] = ref_path
    return ref_paths


def infer_outputs(out_dir: str, fasta: str, fq_unit: FastqUnit):
    """ Infer the paths of the BAM files that are expected to be output
    from the alignment. """
    return [
        path.OneRefAlignmentOutFilePath(
            top=out_dir,
            module=path.Module.ALIGN,
            sample=fq_unit.sample,
            ref=ref,
            ext=path.BAM_EXT
        ).path
        for ref, _ in FastaParser(fasta).parse()
    ]


@docdef.auto()
def run_steps_fq(fq_unit: FastqUnit,
                 fasta: path.RefsetSeqInFilePath | path.OneRefSeqStepFilePath,
                 *,
                 n_procs: int,
                 out_dir: str,
                 temp_dir: str,
                 save_temp: bool,
                 rerun: bool,
                 resume: bool,
                 fastqc: bool,
                 fastqc_extract: bool,
                 cut: bool,
                 cut_q1: int,
                 cut_q2: int,
                 cut_g1: str,
                 cut_a1: str,
                 cut_g2: str,
                 cut_a2: str,
                 cut_o: int,
                 cut_e: float,
                 cut_indels: bool,
                 cut_nextseq: bool,
                 cut_discard_trimmed: bool,
                 cut_discard_untrimmed: bool,
                 cut_m: int,
                 bt2_local: bool,
                 bt2_discordant: bool,
                 bt2_mixed: bool,
                 bt2_dovetail: bool,
                 bt2_contain: bool,
                 bt2_unal: bool,
                 bt2_score_min: str,
                 bt2_i: int,
                 bt2_x: int,
                 bt2_gbar: int,
                 bt2_l: int,
                 bt2_s: str,
                 bt2_d: int,
                 bt2_r: int,
                 bt2_dpad: int,
                 bt2_orient: str,
                 rem_buffer: int) -> list[str]:
    """ Run all steps of alignment for one FASTQ file or one pair of
    mated FASTQ files. """
    # Determine whether alignment needs to be run.
    if not rerun:
        # If alignment is not required to be rerun, then check if all
        # the expected output files already exist.
        expected_output_paths = infer_outputs(out_dir, fasta.path, fq_unit)
        if all(out_path.is_file() for out_path in expected_output_paths):
            # If all the output files already exist, just return them.
            logging.warning(
                f"Skipping alignment for {' and '.join(fq_unit.str_paths)} "
                "because all expected output files already exist. "
                "Add --rerun to rerun.")
            return list(map(str, expected_output_paths))
    # Trim the FASTQ file(s).
    trimmer = FastqTrimmer(top_dir=temp_dir, n_procs=n_procs, fq_unit=fq_unit,
                           save_temp=save_temp, resume=resume)
    if fastqc:
        trimmer.qc(fastqc_extract)
    if cut:
        if resume and all(p.is_file() for p in trimmer.output.paths):
            fq_unit = trimmer.output
        else:
            fq_unit = trimmer.run(cut_q1=cut_q1,
                                  cut_q2=cut_q2,
                                  cut_g1=cut_g1,
                                  cut_a1=cut_a1,
                                  cut_g2=cut_g2,
                                  cut_a2=cut_a2,
                                  cut_o=cut_o,
                                  cut_e=cut_e,
                                  cut_indels=cut_indels,
                                  cut_nextseq=cut_nextseq,
                                  cut_discard_trimmed=cut_discard_trimmed,
                                  cut_discard_untrimmed=cut_discard_untrimmed,
                                  cut_m=cut_m)
            if fastqc:
                trimmed_qc = FastqTrimmer(top_dir=temp_dir,
                                          n_procs=n_procs,
                                          fq_unit=fq_unit,
                                          save_temp=save_temp,
                                          resume=resume)
                trimmed_qc.qc(fastqc_extract)
    # Align the FASTQ to the reference.
    aligner = FastqAligner(top_dir=temp_dir, n_procs=n_procs, fq_unit=fq_unit,
                           fasta=fasta, save_temp=save_temp, resume=resume)
    xam_path = aligner.run(bt2_local=bt2_local,
                           bt2_discordant=bt2_discordant,
                           bt2_mixed=bt2_mixed,
                           bt2_dovetail=bt2_dovetail,
                           bt2_contain=bt2_contain,
                           bt2_unal=bt2_unal,
                           bt2_score_min=bt2_score_min,
                           bt2_i=bt2_i,
                           bt2_x=bt2_x,
                           bt2_gbar=bt2_gbar,
                           bt2_l=bt2_l,
                           bt2_s=bt2_s,
                           bt2_d=bt2_d,
                           bt2_r=bt2_r,
                           bt2_dpad=bt2_dpad,
                           bt2_orient=bt2_orient)
    trimmer.clean()
    # Remove equally mapping reads.
    remover = SamRemoveEqualMappers(top_dir=temp_dir,
                                    n_procs=n_procs,
                                    input_path=xam_path,
                                    save_temp=save_temp,
                                    resume=resume)
    xam_path = remover.run(rem_buffer=rem_buffer)
    aligner.clean()
    # Sort the SAM file and output a BAM file.
    sorter = BamAlignSorter(top_dir=temp_dir,
                            n_procs=n_procs,
                            input_path=xam_path,
                            save_temp=save_temp,
                            resume=resume)
    xam_path = sorter.run()
    remover.clean()
    # Split the BAM file into one file for each reference.
    splitter = BamSplitter(top_dir=temp_dir,
                           n_procs=n_procs,
                           input_path=xam_path,
                           fasta=fasta,
                           save_temp=save_temp,
                           resume=resume)
    bams = splitter.run()
    sorter.clean()
    # Output the BAM files and generate an index for each.
    bams = [BamOutputter(top_dir=out_dir,
                         n_procs=n_procs,
                         input_path=bam,
                         save_temp=save_temp,
                         resume=resume).run()
            for bam in bams]
    splitter.clean()
    return list(map(str, bams))


@docdef.auto()
def run_steps_fqs(fq_units: list[FastqUnit],
                  fasta: str,
                  *,
                  max_procs: int,
                  parallel: bool,
                  out_dir: str,
                  temp_dir: str,
                  save_temp: bool,
                  **kwargs) -> tuple[str, ...]:
    """ Run all steps of alignment for one or more FASTQ files or pairs
    of mated FASTQ files. """
    n_fqs = len(fq_units)
    if n_fqs == 0:
        logging.critical("No FASTQ files were given for alignment.")
        return ()
    # Confirm that there are no duplicate samples and references.
    if dups := check_for_duplicates(fq_units):
        logging.critical(f"Got duplicate samples/refs: {dups}")
        return ()
    if max_procs < 1:
        logging.warning("Maximum CPUs must be ≥ 1: setting to 1")
        max_procs = 1
    refset_path = path.RefsetSeqInFilePath.parse(fasta)
    # Write the temporary FASTA files for demultiplexed FASTQs.
    temp_ref_paths = write_temp_ref_files(temp_dir, fasta, fq_units)
    try:
        # Determine how to parallelize each alignment task.
        n_tasks_parallel, n_procs_per_task = get_num_parallel(n_fqs,
                                                              max_procs,
                                                              parallel,
                                                              hybrid=True)
        # One alignment task will be created for each FASTQ unit.
        # Get the arguments for each task, including n_procs_per_task.
        iter_args = list()
        for fq in fq_units:
            if fq.by_ref:
                # If the FASTQ came from demultiplexing (so contains
                # reads from only one reference), then align to the
                # FASTA of only that reference.
                try:
                    fasta_arg = temp_ref_paths[fq.ref]
                except KeyError:
                    # If the FASTA with that reference does not exist,
                    # then log an error and skip this FASTQ.
                    logging.error(f"Skipped FASTQ(s) {', '.join(fq.str_paths)} "
                                  f"because reference '{fq.ref}' was not found "
                                  f"in FASTA file {fasta}")
                    continue
            else:
                # If the FASTQ may contain reads from ≥ 1 references,
                # then align to the FASTA file with all references.
                fasta_arg = refset_path
            # Add these arguments to the lists of arguments that will be
            # passed to run_steps_fq.
            iter_args.append((fq, fasta_arg))
        # Pass the keyword arguments to every call of run_steps_fq.
        partial_run_steps_fq = partial(run_steps_fq,
                                       **{**dict(n_procs=n_procs_per_task,
                                                 out_dir=out_dir,
                                                 temp_dir=temp_dir,
                                                 save_temp=save_temp),
                                          **kwargs})
        if n_tasks_parallel > 1:
            # Process multiple FASTQ files simultaneously.
            with Pool(n_tasks_parallel) as pool:
                bams = pool.starmap(partial_run_steps_fq, iter_args)
        else:
            # Process the FASTQ files sequentially.
            bams = itertools.starmap(partial_run_steps_fq, iter_args)
    finally:
        if not save_temp:
            # Delete the temporary files before exiting.
            for ref_file in temp_ref_paths.values():
                logging.debug(
                    f"Deleting temporary reference file: {ref_file}")
                ref_file.path.unlink(missing_ok=True)
    # Return a tuple of the final alignment map files.
    return tuple(itertools.chain(*bams))



import os
import numpy as np
import pandas as pd

def run_command(cmd):
    import subprocess
    print(cmd)
    process = subprocess.Popen(cmd.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    return output.decode('utf-8')
                
class RNAstructure(): 
    def __init__(self, rnastructure_path='') -> None:
        self.rnastructure_path = rnastructure_path
        if len(self.rnastructure_path) > 0:
            if self.rnastructure_path[-1] != '/':
                self.rnastructure_path += '/'
        self.directory = 'temp/rnastructure/'

    def fit(self, sequence, reference='reference'):
        self.sequence = sequence
        self.__make_temp_folder()
        self.__make_files()
        self.__create_fasta_file(reference, sequence)


    def predict_ensemble_energy(self):
        cmd = f"{self.rnastructure_path}EnsembleEnergy {self.fasta_file} --sequence"
        splitted_output = run_command(cmd).split(' ')
        return float(splitted_output[splitted_output.index(f"kcal/mol\n\nEnsemble")-1])

    def predict_partition(self, temperature_k =None):
        cmd = f"{self.rnastructure_path}partition {self.fasta_file} {self.pfs_file}"
        if temperature_k != None:
            cmd += ' --temperature '+str(temperature_k)
        run_command(cmd)
        run_command(self.rnastructure_path+'ProbabilityPlot '+ self.pfs_file + ' -t '+self.prob_file)
        with open(self.prob_file,"r") as f:
            lines=f.readlines()
            out={'i':[],'j':[],'p':[]}
            for x in range(len(lines)):
                if x>1:
                    ls=lines[x].split("\t")
                    out["i"]+=[int(ls[0])]
                    out["j"]+=[int(ls[1])]
                    out["p"]+=[float(ls[2])]
        return self.__cast_pairing_prob(out)

    def predict_reference_deltaG(self):
        cmd = f"{self.rnastructure_path}Fold {self.fasta_file} {self.ct_file}"
        run_command(cmd)
        assert os.path.getsize(self.ct_file) != 0, f"{self.ct_file} is empty, check that RNAstructure works"
        return self.__extract_deltaG_struct()

    def __make_temp_folder(self):
        isExist = os.path.exists(self.directory)
        if not isExist:
            os.makedirs(self.directory)
        return self.directory

    def __make_files(self, temp_prefix='temp'):
        self.pfs_file = os.path.join(self.directory, temp_prefix+'.pfs')
        self.ct_file = os.path.join(self.directory, temp_prefix+'.ct')
        self.dot_file = os.path.join(self.directory, temp_prefix+'.dot')
        self.fasta_file = os.path.join(self.directory, temp_prefix+'.fasta')
        self.prob_file = os.path.join(self.directory, temp_prefix+'_prob.txt')

    def __create_fasta_file(self, reference, sequence):
        # push the ref into a temp file
        temp_fasta = open(self.fasta_file, 'w')
        temp_fasta.write('>'+reference+'\n'+sequence)
        temp_fasta.close()

    # cast the temp file into a dot_bracket structure and extract the attributes
    def __extract_deltaG_struct(self):
        run_command(f"ct2dot {self.ct_file} 1 {self.dot_file}")
        temp_dot = open(self.dot_file, 'r')
        first_line = temp_dot.readline().split()
        # If only dots in the structure, no deltaG 
        print('first_line',first_line)
        if len(first_line) == 4:
            _, _, deltaG, _ = first_line
            deltaG = float(deltaG)
        if len(first_line) == 1:
            deltaG, _ = 0.0, first_line[0][1:]

        sequence = temp_dot.readline()[:-1] #  Remove the \n
        structure = temp_dot.readline()[:-1] # Remove the \n
        return deltaG, structure

    def run(self, sequence):
        if os.path.exists(os.path.join(self.directory, sequence+'.txt')):
            # read from file
            with open(os.path.join(self.directory, sequence+'.txt'), 'r') as f:
                deltaG, structure = f.readline().split()
        else:
            self.fit(sequence)
            deltaG, structure = self.predict_reference_deltaG()
            with open(os.path.join(self.directory, sequence+'.txt'), 'w') as f:
                f.write(f"{deltaG} {structure}")
        if deltaG == 'void': 
            deltaG = 0.0
        else:
            deltaG = float(deltaG)
        return {'deltaG':deltaG, 'structure':structure}

if __name__ == "__main__":
    rna = RNAstructure('/Users/ymdt/src/RNAstructure/exe/')
    rna.fit(sequence='AAGATATTCGAAACCACTCGATCGACTAGCATCAGCTGACTAGCTAGCATGCATCAAGAATATCTT')
    print("DeltaG + structure:", rna.predict_reference_deltaG())
    print("Ens. energy:", rna.predict_ensemble_energy())
    print("One line command:", rna.run('AAGATATTCGAAACCACTCGATCGACTAGCATCAGCTGACTAGCTAGCATGCATCAAGAATATCTT'))


from datetime import datetime
from enum import Enum, IntEnum
import logging
import os
from typing import Any, Callable, Iterable

from click import Choice, Option, Parameter, Path

from .seq import (MATCH_INT, DELET_INT, INS_5_INT, INS_3_INT,
                  SUB_N_INT, SUB_A_INT, SUB_C_INT, SUB_G_INT, SUB_T_INT,
                  AMBIG_INT)

# System information
CWD = os.getcwd()
if (NUM_CPUS := os.cpu_count()) is None:
    logging.warning("Failed to determine CPU count: defaulting to 1")
    NUM_CPUS = 1

DEFAULT_PHRED_ENC = 33
DEFAULT_MIN_PHRED = 25


class DreemCommandName(Enum):
    """ Commands for DREEM """
    TEST = "test"
    DEMULTIPLEX = "demultiplex"
    ALIGN = "align"
    VECTOR = "vector"
    CLUSTER = "cluster"
    AGGREGATE = "aggregate"
    DRAW = "draw"


class MateOrientationOption(Enum):
    """ Options of mate orientation for alignment with Bowtie2.

    See Bowtie2 manual for full documentation:
    https://bowtie-bio.sourceforge.net/bowtie2/manual.shtml
    """
    FR = "fr"
    RF = "rf"
    FF = "ff"


class CountOption(Enum):
    """ Options for count-based statistics """
    COVERAGE = "v"
    MATCHES = "w"
    DELETIONS = "d"
    ALL_INSERTIONS = "i"
    INSERTIONS_5PRIME = "i5"
    INSERTIONS_3PRIME = "i3"
    ALL_SUBSTITUTIONS = "s"
    SUBS_TO_ADENINE = "a"
    SUBS_TO_CYTOSINE = "c"
    SUBS_TO_GUANINE = "g"
    SUBS_TO_THYMINE = "t"
    ALL_MUTATIONS = "m"


class CountOptionValue(IntEnum):
    """ Values for count-based statistics """
    COVERAGE = AMBIG_INT
    MATCHES = MATCH_INT
    DELETIONS = DELET_INT
    ALL_INSERTIONS = INS_5_INT + INS_3_INT
    INSERTIONS_5PRIME = INS_5_INT
    INSERTIONS_3PRIME = INS_3_INT
    ALL_SUBSTITUTIONS = SUB_N_INT
    SUBS_TO_ADENINE = SUB_A_INT
    SUBS_TO_CYTOSINE = SUB_C_INT
    SUBS_TO_GUANINE = SUB_G_INT
    SUBS_TO_THYMINE = SUB_T_INT
    ALL_MUTATIONS = AMBIG_INT - MATCH_INT


class AdapterSequence(Enum):
    """ Adapter sequences """
    ILLUMINA_3P = "AGATCGGAAGAGC"


# Input/output options
opt_out_dir = Option(("--out-dir", "-o"),
                     type=Path(file_okay=False),
                     default=os.path.join(CWD, "output"),
                     help="Where to output all finished files")
opt_temp_dir = Option(("--temp-dir", "-t"),
                      type=Path(file_okay=False),
                      default=os.path.join(CWD, "temp"),
                      help="Where to write all temporary files")
opt_save_temp = Option(("--save-temp/--erase-temp",),
                       type=bool,
                       default=False,
                       help=("Whether to save or erase temporary files "
                             "after the program exits"))

# Resource usage options
opt_parallel = Option(("--parallel/--no-parallel",),
                      type=bool,
                      default=True,
                      help="Whether to run multiple jobs in parallel")
opt_max_procs = Option(("--max-procs",),
                       type=int,
                       default=NUM_CPUS,
                       help="Maximum number of simultaneous processes")

# Experiment and analysis setup options
opt_library = Option(("--library",),
                     type=Path(dir_okay=False),
                     default="",
                     help="Library CSV file")
opt_samples = Option(("--samples",),
                     type=Path(dir_okay=False),
                     default="",
                     help="Samples file")
opt_rerun = Option(("--rerun/--no-rerun",),
                   type=bool,
                   default=False,
                   help="Whether to regenerate files that already exist")
opt_resume = Option(("--resume/--no-resume",),
                    type=bool,
                    default=False,
                    help=("Whether to use any existing temporary files "
                          "to resume a process that terminated"))

# Reference sequence (FASTA) files
opt_fasta = Option(("--fasta", "-f"),
                   type=Path(exists=True, dir_okay=False),
                   help="FASTA file of all reference sequences in the project")

# Sequencing read (FASTQ) files
opt_fastqs = Option(("--fastqs", "-s"),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ file of single-end reads")
opt_fastqi = Option(("--fastqi", "-i"),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ file of interleaved paired reads")
opt_fastq1 = Option(("--fastq1", "-1"),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ file of mate 1 paired-end reads")
opt_fastq2 = Option(("--fastq2", "-2"),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=(),
                    help="FASTQ file of mate 2 paired-end reads")

# Sequencing read (FASTQ/BAM) options
opt_phred_enc = Option(("--phred-enc", "-e"),
                       type=int,
                       default=DEFAULT_PHRED_ENC,
                       help="Phred score encoding in FASTQ/SAM/BAM files")
opt_min_phred = Option(("--min-phred", "-q"),
                       type=int,
                       default=DEFAULT_MIN_PHRED,
                       help="Minimum Phred score to use a base call")
opt_fastqc = Option(("--fastqc/--no-fastqc",),
                    type=bool,
                    default=True,
                    help="Whether to check quality of FASTQ files")
opt_fastqc_extract = Option(("--fastqc-extract/--fastqc-no-extract",),
                            type=bool,
                            default=True,
                            help="Whether to unzip FASTQC reports")

# Demultiplexing options
opt_demultiplex = Option(("--demult-on/--demult-off",),
                         type=bool,
                         default=False,
                         help="Whether to run demultiplexing")
opt_barcode_start = Option(("--barcode-start",),
                           type=int,
                           default=0)
opt_barcode_length = Option(("--barcode-length",),
                            type=int,
                            default=0)
opt_max_barcode_mismatches = Option(("--max_barcode_mismatches",),
                                    type=int,
                                    default=1)

# Demultiplexed sequencing read (FASTQ) directories
opt_fastqs_dir = Option(("--fastqs-dir", "-S"),
                        type=Path(exists=True, file_okay=False),
                        multiple=True,
                        default=(),
                        help="Demultiplexed FASTQ files of single-end reads")
opt_fastqi_dir = Option(("--fastqi-dir", "-I"),
                        type=Path(exists=True, file_okay=False),
                        multiple=True,
                        default=(),
                        help="Demultiplexed FASTQ files of interleaved paired-end reads")
opt_fastq12_dir = Option(("--fastq12-dir", "-P"),
                         type=Path(exists=True, file_okay=False),
                         multiple=True,
                         default=(),
                         help="Demultiplexed pairs of FASTQ files of mate 1 and mate 2 reads")

# Alignment map (BAM) files
opt_bamf = Option(("--bamf", "-b"),
                  type=Path(exists=True, dir_okay=False),
                  multiple=True,
                  default=(),
                  help="BAM file")
opt_bamd = Option(("--bamd", "-B"),
                  type=Path(exists=True, file_okay=False),
                  multiple=True,
                  default=(),
                  help="Directory of BAM files")

# Adapter trimming options with Cutadapt
opt_cutadapt = Option(("--cut/--no-cut",),
                      type=bool,
                      default=True,
                      help="Whether to trim reads with Cutadapt before alignment")
opt_cut_q1 = Option(("--cut-q1",),
                    type=int,
                    default=DEFAULT_MIN_PHRED,
                    help="Phred score for read 1 quality trimming")
opt_cut_q2 = Option(("--cut-q2",),
                    type=int,
                    default=DEFAULT_MIN_PHRED,
                    help="Phred score for read 2 quality trimming")
opt_cut_g1 = Option(("--cut-g1",),
                    type=str,
                    multiple=True,
                    default=(),
                    help="5' adapter for read 1")
opt_cut_a1 = Option(("--cut-a1",),
                    type=str,
                    multiple=True,
                    default=(AdapterSequence.ILLUMINA_3P.value,),
                    help="3' adapter for read 1")
opt_cut_g2 = Option(("--cut-g2",),
                    type=str,
                    multiple=True,
                    default=(),
                    help="5' adapter for read 2")
opt_cut_a2 = Option(("--cut-a2",),
                    type=str,
                    multiple=True,
                    default=(AdapterSequence.ILLUMINA_3P.value,),
                    help="3' adapter for read 2")
opt_cut_o = Option(("--cut-O",),
                   type=int,
                   default=6,
                   help="Minimum overlap of read and adapter")
opt_cut_e = Option(("--cut-e",),
                   type=float,
                   default=0.1,
                   help="Error tolerance for adapters")
opt_cut_indels = Option(("--cut-indels/--cut-no-indels",),
                        type=bool,
                        default=True,
                        help="Whether to allow indels in adapters")
opt_cut_nextseq = Option(("--cut-nextseq/--cut-no-nextseq",),
                         type=bool,
                         default=False,
                         help="Whether to trim high-quality Gs from 3' end")
opt_cut_discard_trimmed = Option(
    ("--cut-discard-trimmed/--cut-keep-trimmed",),
    type=bool,
    default=False,
    help="Whether to discard reads in which an adapter was found")
opt_cut_discard_untrimmed = Option(
    ("--cut-discard-untrimmed/--cut-keep-untrimmed",),
    type=bool,
    default=False,
    help="Whether to discard reads in which no adapter was found")
opt_cut_m = Option(("--cut-m",),
                   type=int,
                   default=20,
                   help="Discard reads shorter than this length after trimming")

# Alignment options with Bowtie2
opt_bt2_local = Option(("--bt2-local/--bt2-end-to-end",),
                       type=bool,
                       default=True,
                       help="Whether to perform local or end-to-end alignment")
opt_bt2_discordant = Option(("--bt2-discordant/--bt2-no-discordant",),
                            type=bool,
                            default=False,
                            help="Whether to output discordant alignments")
opt_bt2_mixed = Option(("--bt2-mixed/--bt2-no-mixed",),
                       type=bool,
                       default=False,
                       help="Whether to align individual mates of unaligned pairs")
opt_bt2_dovetail = Option(("--bt2-dovetail/--bt2-no-dovetail",),
                          type=bool,
                          default=False,
                          help="Whether to treat dovetailed mate pairs as concordant")
opt_bt2_contain = Option(("--bt2-contain/--bt2-no-contain",),
                         type=bool,
                         default=True,
                         help="Whether to treat nested mate pairs as concordant")
opt_bt2_unal = Option(("--bt2-unal/--bt2-no-unal",),
                      type=bool,
                      default=False,
                      help="Whether to output unaligned reads")
opt_bt2_i = Option(("--bt2-I",),
                   type=int,
                   default=0,
                   help="Minimum fragment length for valid paired-end alignments")
opt_bt2_x = Option(("--bt2-X",),
                   type=int,
                   default=600,
                   help="Maximum fragment length for valid paired-end alignments")
opt_bt2_score_min = Option(("--bt2-score-min",),
                           type=str,
                           default="L,0,0.5",
                           help="Minimum score for a valid alignment")
opt_bt2_s = Option(("--bt2-i", "bt2_s"),
                   type=str,
                   default="L,1,0.1",
                   help="Seed interval")
opt_bt2_l = Option(("--bt2-L",),
                   type=int,
                   default=12,
                   help="Seed length")
opt_bt2_gbar = Option(("--bt2-gbar",),
                      type=int,
                      default=4,
                      help="Minimum distance of a gap from end of a read")
opt_bt2_d = Option(("--bt2-D",),
                   type=int,
                   default=4,
                   help="Maximum number of failed seed extensions")
opt_bt2_r = Option(("--bt2-R",),
                   type=int,
                   default=2,
                   help="Maximum number of times to re-seed")
opt_bt2_dpad = Option(("--bt2-dpad",),
                      type=int,
                      default=2,
                      help="Width of padding on alignment matrix, to allow gaps")
opt_bt2_orient = Option(("--bt2-orient",),
                        type=Choice(tuple(op.value for op
                                          in MateOrientationOption),
                                    case_sensitive=False),
                        default=MateOrientationOption.FR.value,
                        help="Valid orientations of paired-end mates")

# Other alignment options
opt_rem_buffer = Option(("--rem-buffer",),
                        type=int,
                        default=65536,
                        help=("Maximum number of reads to hold in memory when "
                              "removing reads with multiple equal alignments"))

# Reference region specification options
opt_coords = Option(("--coords", "-c"),
                    type=(str, int, int),
                    multiple=True,
                    default=(),
                    help=("Reference name, 5' end, and 3' end of a region; "
                          "coordinates are 1-indexed and include both ends"))
opt_primers = Option(("--primers", "-p"),
                     type=(str, str, str),
                     multiple=True,
                     default=(),
                     help=("Reference name, forward primer, and reverse primer "
                           "of a region; reverse primer must be given 5' to 3'"))
opt_primer_gap = Option(("--primer-gap",),
                        type=int,
                        default=2,
                        help=("Number of bases to leave as a gap between the "
                              "end of a primer and the end of the region"))
opt_cfill = Option(("--cfill/--no-cfill",),
                   type=bool,
                   default=False,
                   help=("Whether, for every reference that was not explicitly "
                         "given at least one region (by --initial_coords or "
                         "--primers), to generate coordinates covering the "
                         "entire reference sequence automatically"))

# Vectoring options
opt_batch_size = Option(("--batch-size", "-z"),
                        type=float,
                        default=32.0,
                        help=("Maximum size of each batch of mut_vectors, "
                              "in millions of base calls"))
opt_strict_pairs = Option(("--strict-pairs/--no-strict-pairs",),
                          type=bool,
                          default=True,
                          help=("Whether to require that every paired "
                                "read that maps to a region also have "
                                "a mate that maps to the region"))
opt_ambid = Option(("--ambid/--no-ambid",),
                   type=bool,
                   default=True,
                   help=("Whether to find and label all ambiguous "
                         "insertions and deletions (improves accuracy "
                         "but runs slower)"))

# Mutational profile report files
opt_report = Option(("--mp-report", "-r"),
                    type=Path(exists=True, dir_okay=False),
                    multiple=True,
                    default=())

# Clustering options
opt_cluster = Option(("--cluster-on/--cluster-off",),
                     type=bool,
                     default=False,
                     help="Whether to run clustering")
opt_max_clusters = Option(("--max-clusters", "-k"),
                          type=int,
                          default=3)
opt_min_iter = Option(("--min-iter", "-m"),
                      type=int,
                      default=100)
opt_max_iter = Option(("--max-iter", "-x"),
                      type=int,
                      default=500)
opt_signal_thresh = Option(("--signal-thresh", "-s"),
                           type=float,
                           default=0.005)
opt_info_thresh = Option(("--info-thresh", "-i"),
                         type=float,
                         default=0.05)
opt_include_gu = Option(("--include-gu/--exclude-gu",),
                        type=bool,
                        default=False)
opt_include_del = Option(("--include-del/--exclude-del",),
                         type=bool,
                         default=False)
opt_min_reads = Option(("--min-reads", "-v"),
                       type=int,
                       default=1000)
opt_convergence_cutoff = Option(("--convergence-cutoff", "-g"),
                                type=float,
                                default=0.5)
opt_num_runs = Option(("--num-runs", "-n"),
                      type=int,
                      default=10)

# Statistics options
opt_stats_count = Option(("--count", "-c"),
                         type=Choice(tuple(CountOption),
                                     case_sensitive=False),
                         multiple=True,
                         default=(), )
opt_stats_frac = Option(("--frac", "-f"),
                        type=Choice(tuple(CountOption),
                                    case_sensitive=False),
                        multiple=True,
                        default=(), )

# Aggregation
RNASTRUCTURE_PATH = ''
RNASTRUCTURE_TEMPERATURE = False
RNASTRUCTURE_FOLD_ARGS = ''
RNASTRUCTURE_DMS = False
RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE = 0.04
RNASTRUCTURE_DMS_MAX_PAIRED_VALUE = 0.01
RNASTRUCTURE_PARTITION = False
RNASTRUCTURE_PROBABILITY = False

sample = Option(("--sample", "-s"),
                type=str,
                default='output')

clustering_file = Option(("--clustering_file", "-cf"),
                            type=Path(exists=True, dir_okay=False),
                            default='')

rnastructure_path = Option(("--rnastructure_path", "-rs"),
                           type=Path(exists=True))
rnastructure_temperature = Option(("--rnastructure_temperature", "-rst"),
                                  type=bool, default=False)
rnastructure_fold_args = Option(("--rnastructure_fold_args", "-rsa"),
                                type=str,
                                default=RNASTRUCTURE_FOLD_ARGS )
rnastructure_dms = Option(("--rnastructure_dms", "-rsd"),
                          type=bool,
                          default=RNASTRUCTURE_DMS,
                          help="Use the DMS signal to make predictions with RNAstructure")
rnastructure_dms_min_unpaired_value = Option(("--rnastructure_dms_min_unpaired_value", "-rsdmin"),
                                             type=int,
                                             default=RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE,
                                             help="Minimum unpaired value for using the dms signal as an input for RNAstructure")
rnastructure_dms_max_paired_value = Option(("--rnastructure_dms_max_paired_value", "-rsdmax"),
                                           type=int,
                                           default=RNASTRUCTURE_DMS_MAX_PAIRED_VALUE,
                                           help="Maximum paired value for using the dms signal as an input for RNAstructure")
rnastructure_partition = Option(("--rnastructure_partition", "-rspa"),
                                type=bool,
                                default=RNASTRUCTURE_PARTITION,
                                help="Use RNAstructure partition function to predict free energy")
rnastructure_probability = Option(("--rnastructure_probability", "-rspr"),
                                  type=bool,
                                  default=RNASTRUCTURE_PROBABILITY,
                                  help="Use RNAstructure partition function to predict per-base mutation probability")

# Logging options
opt_verbose = Option(("--verbose", "-v"),
                     count=True)
opt_quiet = Option(("--quiet", "-q"),
                   count=True)
opt_logfile = Option(("--log",),
                     type=Path(exists=False, dir_okay=False),
                     default=os.path.join(CWD, datetime.now().strftime(
                         "dreem-log_%Y-%m-%d_%H:%M:%S"
                     )))


def merge_params(*param_lists: list[Parameter]):
    """ Merge lists of Click parameters, dropping duplicates. """
    params = list()
    names = set()
    for param_list in param_lists:
        for param in param_list:
            if param.name not in names:
                params.append(param)
                names.add(param.name)
    return params


class DreemCommand(object):
    """

    """

    def __init__(self,
                 cli_func: Callable,
                 imports: tuple[str, ...],
                 exports: tuple[str, ...],
                 result_key: str | None):
        """
        Parameters
        ----------
        cli_func: callable
            Command line function to wrap
        imports: tuple[str]
            Key(s) to import from the context object into ```kwargs```
            before calling ```cli_func(**kwargs)```. Imported keys
            override existing keys in ```kwargs```.
        exports: tuple[str]
            Key(s) to export from ```kwargs``` to the context object.
            Exported keys override existing keys in ```ctx_obj```.
        result_key: str | None
            Key under which to export the result of the function, or
            None to discard the result.
        """
        self._cli_func = cli_func
        self._exports = exports
        self._imports = imports
        self._result_key = result_key

    @staticmethod
    def _update_keys(updated: dict[str, Any],
                     updater: dict[str, Any],
                     keys: Iterable[str]):
        """ For every key in ```keys```, if ```updater``` contains the
        key, then add its key-value pair to ```updated```, overriding
        any existing key-value pairs with the same key names in
        ```updated```; otherwise, ignore the key. """
        for key in keys:
            try:
                updated[key] = updater[key]
            except KeyError:
                pass

    def __call__(self, ctx_obj: dict[str, Any], *args: Any, **kwargs: Any):
        """
        Wrapper that calls ```cli_func(*args, **kwargs)``` after adding
        keyword arguments from ```ctx_obj``` to ```kwargs```.

        Parameters
        ----------
        ctx_obj: dict[str, Any]
            Object storing attributes of the Click context
        *args: Any:
            Positional arguments to pass to ```cli_func```
        **kwargs: Any:
            Keyword arguments to pass to ```cli_func```

        Return
        ------
        Any
            Return value of ```cli_func(**kwargs)```
        """
        # Make shallow copy of kwargs so the outer scope can assume that
        # kwargs is not modified.
        kwargs = kwargs.copy()
        # Import selected keyword arguments from the context object into
        # the dictionary of keyword arguments.
        self._update_keys(kwargs, ctx_obj, self._imports)
        # Call the function and optionally store its return value(s).
        result = self._cli_func(*args, **kwargs)
        if self._result_key is not None:
            kwargs[self._result_key] = result
        # Export all keyword arguments to the context object so that
        # subsequent commands can import the values.
        self._update_keys(ctx_obj, kwargs, self._exports)
        return result


def dreem_command(imports: tuple[str, ...] = (),
                  exports: tuple[str, ...] = (),
                  result_key: str | None = None):
    def command_decorator(func: Callable):
        return DreemCommand(func, imports, exports, result_key)

    return command_decorator



"""
Logging module of DREEM

Purpose
-------
Central manager of logging.
"""

import logging


MAX_VERBOSE = 2
MAX_QUIET = 2


DREEM_BUG_TEXT = ("This error indicates a bug in DREEM itself, not a problem "
                  "with the files or settings you have used to run DREEM. To "
                  "report this bug so that it can be fixed, please visit "
                  "https://github.com/rouskinlab/dreem/issues, click on "
                  "'New Issue', and describe what you did and what the error "
                  "message was. The more details you include, the easier it "
                  "will be to identify and fix the bug. We apologize for this "
                  "inconvenience.")


def set_verbosity(verbose: int, quiet: int):
    """ Set the logging level based on the verbose and quiet arguments.

    Parameters
    ----------
    verbose: int [0, 2]
        0 (): Log only warnings and errors
        1 (-v): Also log status updates
        2 (-vv): Also log detailed information (useful for debugging)
    quiet: int [0, 2]
        0 (): Suppress only status updates and detailed information
        1 (-q): Also suppress warnings
        2 (-qq): Also suppress non-critical error messages (discouraged)

    Giving both ```verbose``` and ```quiet``` flags causes the verbosity
    to default to ```verbose=0```, ```quiet=0```.
    """

    # Limit verbose and quiet to 2.
    if verbose > MAX_VERBOSE:
        logging.warning(f"Setting 'verbose' to {MAX_VERBOSE} (got {verbose})")
        verbose = MAX_VERBOSE
    if quiet > MAX_QUIET:
        logging.warning(f"Setting 'quiet' to {MAX_QUIET} (got {quiet})")
        quiet = MAX_QUIET

    # Set logging level based on verbose and quiet.
    if verbose == 0 and quiet == 2:
        level = logging.CRITICAL
    elif verbose == 0 and quiet == 1:
        level = logging.ERROR
    elif verbose == 0 and quiet == 0:
        level = logging.WARNING
    elif verbose == 1 and quiet == 0:
        level = logging.INFO
    elif verbose == 2 and quiet == 0:
        level = logging.DEBUG
    else:
        logging.warning(f"Invalid options: verbose={verbose}, quiet={quiet}. "
                        "Defaulting to verbose=0, quiet=0.")
        level = logging.WARNING
    logging.basicConfig(level=level)
    return level




import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import os
import pytest
from dreem.main import run
from dreem.util.files_sanity import compare_fields
import json

sample = 'my_test_sample'
test_files = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), 'test_files')
expected_file = json.load(open(os.path.join(test_files, '{}.json'.format(sample)), 'r'))
top_dir = test_files.replace('test_files', 'test_output')
temp_dir = test_files.replace('test_files', 'test_temp')
output_file_path = os.path.join(top_dir, '{}.json'.format(sample))

FIELDS_FROM_ALIGNMENT_REPORT = ['num_reads','skips_low_mapq','skips_short_reads','skips_too_many_muts']

def output_file():
    return json.load(open(output_file_path, 'r'))

def references_from_json(json_file):
    return [k for k in json_file.keys() if type(json_file[k]) == dict]

def attribute_from_reference(json_file):
    reference = references_from_json(json_file)[0]
    return json_file[reference].keys()

def section_from_reference(json_file):
    reference = references_from_json(json_file)[0]
    return [k for k in json_file[reference].keys() if type(json_file[reference][k]) == dict]

def attribute_from_section(json_file):
    reference = references_from_json(json_file)[0]
    section = section_from_reference(json_file)[0]
    return json_file[reference][section].keys()

def attribute_from_pop_avg(json_file):
    reference = references_from_json(json_file)[0]
    section = section_from_reference(json_file)[0]
    return json_file[reference][section]['pop_avg'].keys()

def test_run():        
    run(
        out_dir= top_dir,      
        temp_dir = temp_dir,
        fastq1 = ('{}/{}_R1.fastq'.format(test_files,sample),),
        fastq2 = ('{}/{}_R2.fastq'.format(test_files,sample),),
        fasta = '{}/{}.fasta'.format(test_files,sample),
        library = '{}/library.csv'.format(test_files,sample),
        samples = '{}/samples.csv'.format(test_files,sample),
        rerun=True,
        rnastructure_path = '/Users/ymdt/src/RNAstructure/exe',
        )
        
def test_output_exists():        
    assert os.path.exists(os.path.join(os.getcwd(),'test_output','my_test_sample.json'))

@pytest.mark.parametrize('attr', expected_file.keys())
def test_sample_attributes(attr):
    if not type(expected_file[attr]) == dict:
        compare_fields(expected_file, output_file(), [attr])

@pytest.mark.parametrize('reference', references_from_json(expected_file))
@pytest.mark.parametrize('attr', attribute_from_reference(expected_file))
def test_library_attributes(reference,attr):
    if not type(expected_file[reference][attr]) == dict and not attr in FIELDS_FROM_ALIGNMENT_REPORT:
        compare_fields(expected_file, output_file(), [reference, attr])

@pytest.mark.parametrize('reference,section', [(c,s) for c in list(references_from_json(expected_file)) for s in section_from_reference(expected_file)])
@pytest.mark.parametrize('attr', attribute_from_section(expected_file))
def test_sections_attributes(reference,section,attr):
    if not type(expected_file[reference][section][attr]) == dict and not attr in FIELDS_FROM_ALIGNMENT_REPORT:
        compare_fields(expected_file, output_file(), [reference, section, attr])

@pytest.mark.parametrize('reference', references_from_json(expected_file))
@pytest.mark.parametrize('attr', attribute_from_pop_avg(expected_file) ) 
def test_mp_pop_avg(reference,attr):
    for section in section_from_reference(expected_file):
        compare_fields(expected_file, output_file(), [reference, section, 'pop_avg', attr])

@pytest.mark.parametrize('reference', references_from_json(expected_file))
def test_section_idx(reference):
    output = output_file()
    for section in section_from_reference(expected_file):
        ss, se = output[reference][section]['section_start'], output[reference][section]['section_end']
        assert len(output[reference][section]['sequence']) == se-ss+1, 'section length is not correct: {} != {}-{}+1'.format(len(output[reference][section]['sequence']), se, ss)
        assert output[reference]['sequence'].index(output[reference][section]['sequence']) == ss-1, 'section start is not correct'
        assert output[reference]['sequence'][ss-1:se] == output[reference][section]['sequence'], 'section sequence is not correct'
        
if __name__ == '__main__':
    # remove test files
    os.system('rm {}'.format(output_file_path))
    test_run()



import logging
import os
from click import Context, group, pass_context

from . import align, cluster, demultiplex, test, vector, aggregate
from .util import docdef
from .util.cli import (merge_params, opt_demultiplex, opt_cluster, opt_quiet,
                       opt_verbose)
from .util.logio import set_verbosity
from .util import path

verbose_params = [
    opt_verbose,
    opt_quiet,
]

all_params = merge_params(verbose_params,
                          [opt_demultiplex],
                          # demultiplex.params,
                          align.params,
                          vector.params,
                          [opt_cluster],
                          # cluster.params,
                          # aggregate.params,
                          # draw.params
                          )


# Group for all DREEM commands
@group(params=all_params, 
       chain=True,
       invoke_without_command=True,
       context_settings={"show_default": True})
@pass_context
def cli(ctx: Context, verbose: int, quiet: int, **kwargs):
    """ DREEM command line interface """
    # Set verbosity level for logging.
    set_verbosity(verbose, quiet)
    # Ensure context object exists and is a dict.
    ctx.ensure_object(dict)
    # If no subcommand was given, then run the entire pipeline.
    if ctx.invoked_subcommand is None:
        run(**kwargs)
 

# Add all commands to the DREEM CLI command group.
#cli.add_command(test.cli)
cli.add_command(demultiplex.cli)
cli.add_command(align.cli)
cli.add_command(vector.cli)
cli.add_command(cluster.cli)


@docdef.auto()
def run(*,
        # General options
        out_dir: str,
        temp_dir: str,
        save_temp: bool,
        rerun: bool,
        resume: bool,
        max_procs: int,
        parallel: bool,
        library: str,
        fasta: str,
        fastqs: tuple[str],
        fastqi: tuple[str],
        fastq1: tuple[str],
        fastq2: tuple[str],
        phred_enc: int,
        # Demultiplexing options
        demult_on: bool,
        # FIXME: add parameters for demultiplexing
        # Alignment options
        fastqs_dir: tuple[str],
        fastqi_dir: tuple[str],
        fastq12_dir: tuple[str],
        fastqc: bool,
        fastqc_extract: bool,
        cut: bool,
        cut_q1: int,
        cut_q2: int,
        cut_g1: str,
        cut_a1: str,
        cut_g2: str,
        cut_a2: str,
        cut_o: int,
        cut_e: float,
        cut_indels: bool,
        cut_nextseq: bool,
        cut_discard_trimmed: bool,
        cut_discard_untrimmed: bool,
        cut_m: int,
        bt2_local: bool,
        bt2_discordant: bool,
        bt2_mixed: bool,
        bt2_dovetail: bool,
        bt2_contain: bool,
        bt2_unal: bool,
        bt2_score_min: str,
        bt2_i: int,
        bt2_x: int,
        bt2_gbar: int,
        bt2_l: int,
        bt2_s: str,
        bt2_d: int,
        bt2_r: int,
        bt2_dpad: int,
        bt2_orient: str,
        rem_buffer: int,
        # Vectoring
        bamf: tuple[str],
        bamd: tuple[str],
        coords: tuple[tuple[str, int, int], ...],
        primers: tuple[tuple[str, str, str], ...],
        primer_gap: int,
        cfill: bool,
        min_phred: int,
        ambid: bool,
        strict_pairs: bool,
        batch_size: float,
        # Clustering
        cluster_on: bool,
        # Aggregate
        samples: str,
        rnastructure_path: str,
        ):
    """ Run entire DREEM pipeline. """
    # Demultiplexing
    if demult_on:
        fastqs_dir_dm, fastqi_dir_dm, fastq12_dir_dm = demultiplex.run(
            # FIXME: add arguments
        )
        fastqs = ()
        fastqi = ()
        fastq1 = ()
        fastq2 = ()
        fastqs_dir = fastqs_dir + fastqs_dir_dm
        fastqi_dir = fastqi_dir + fastqi_dir_dm
        fastq12_dir = fastq12_dir + fastq12_dir_dm
    # Alignment
    bams_aln = align.run(
        out_dir=out_dir,
        temp_dir=temp_dir,
        save_temp=save_temp,
        rerun=rerun,
        resume=resume,
        max_procs=max_procs,
        parallel=parallel,
        fasta=fasta,
        fastqs=fastqs,
        fastqi=fastqi,
        fastq1=fastq1,
        fastq2=fastq2,
        fastqs_dir=fastqs_dir,
        fastqi_dir=fastqi_dir,
        fastq12_dir=fastq12_dir,
        phred_enc=phred_enc,
        fastqc=fastqc,
        fastqc_extract=fastqc_extract,
        cut=cut,
        cut_q1=cut_q1,
        cut_q2=cut_q2,
        cut_g1=cut_g1,
        cut_a1=cut_a1,
        cut_g2=cut_g2,
        cut_a2=cut_a2,
        cut_o=cut_o,
        cut_e=cut_e,
        cut_indels=cut_indels,
        cut_nextseq=cut_nextseq,
        cut_discard_trimmed=cut_discard_trimmed,
        cut_discard_untrimmed=cut_discard_untrimmed,
        cut_m=cut_m,
        bt2_local=bt2_local,
        bt2_discordant=bt2_discordant,
        bt2_mixed=bt2_mixed,
        bt2_dovetail=bt2_dovetail,
        bt2_contain=bt2_contain,
        bt2_unal=bt2_unal,
        bt2_score_min=bt2_score_min,
        bt2_i=bt2_i,
        bt2_x=bt2_x,
        bt2_gbar=bt2_gbar,
        bt2_l=bt2_l,
        bt2_s=bt2_s,
        bt2_d=bt2_d,
        bt2_r=bt2_r,
        bt2_dpad=bt2_dpad,
        bt2_orient=bt2_orient,
        rem_buffer=rem_buffer,
    )
    bamf = bamf + bams_aln
    # Vectoring
    profiles_vec = vector.run(
        out_dir=out_dir,
        temp_dir=temp_dir,
        save_temp=save_temp,
        rerun=rerun,
        resume=resume,
        max_procs=max_procs,
        parallel=parallel,
        fasta=fasta,
        bamf=bamf,
        bamd=bamd,
        library=library,
        cfill=cfill,
        coords=coords,
        primers=primers,
        primer_gap=primer_gap,
        phred_enc=phred_enc,
        min_phred=min_phred,
        ambid=ambid,
        strict_pairs=strict_pairs,
        batch_size=batch_size,
    )
    if cluster_on:
        cluster_results = cluster.run(
            # FIXME: add arguments
        )
    # Aggregate
    # TODO: make sample better
    sample = [f.split("/")[-1].split("_R1.f")[0] for f in fastq1]
    for s in sample:
        aggregate.main.run(
            out_dir=out_dir,
            fasta=fasta,
            library=library,
            samples=samples,
            clustering_file = cluster_results if cluster_on else None,
            bv_files = profiles_vec,
            sample = s,
            rnastructure_path = rnastructure_path,
        )

if __name__ == "__main__":
    cli()



from matplotlib import pyplot as plt
import numpy as np
import pandas as pd

from ..util.cli import CountOption, CountOptionValue
from ..util.seq import AMBIG_INT
from ..vector.profile import VectorReader, trans_vectors_block



def run(out_dir: str,
        report: tuple[str],
        count: tuple[str] = (),
        frac: tuple[str] = ()):
    for rep in report:
        reader = VectorReader.load(rep)
        with open(f"test_vectors.txt", "wb") as f:
            f.write(trans_vectors_block(reader.get_all_vectors()))

        for query in count:
            quint = CountOptionValue[CountOption(query).name]
            colors = [{"A": "red", "C": "blue", "G": "orange", "T": "green"}[reader.ref_seq.decode()[pos-1]]
                      for pos in reader.positions]
            counts = reader.count_query(quint)
            plt.bar(counts.index, counts, color=colors)
            plt.show()
        for query in frac:
            quint = CountOptionValue[CountOption(query).name]
            colors = [{"A": "red", "C": "blue", "G": "orange", "T": "green"}[reader.ref_seq.decode()[pos-1]]
                      for pos in reader.positions]
            counts = reader.count_query(quint)
            xcounts = reader.count_query(AMBIG_INT ^ quint)
            f = counts / (counts + xcounts)
            plt.bar(f.index, f, color=colors)
            plt.show()



import os, random, sys
import pandas as pd
import numpy as np
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
from dreem.util.rnastructure import RNAstructure
from dreem.draw import Study
import json

test_file_folder = os.path.join(os.path.dirname((os.path.dirname(os.path.dirname(__file__)))), 'test_files')

def create_sequence(L):
    return ''.join([random.choice('ACGT') for _ in range(L)])

def invert_sequence(seq):
    return ''.join([{'A':'T','T':'A','C':'G','G':'C'}[s] for s in seq])[::-1]

def next_base(base):
    return {'A':'C','C':'G','G':'T','T':'A',0:1}[base]

def prev_base(base):
    return {'A':'T','C':'A','G':'C','T':'G',0:1}[base]

def hamming_distance(s1, s2):
    return sum(c1 != c2 for c1, c2 in zip(s1, s2))

def print_fasta_line(f, id, seq):
    f.write('>{}\n{}\n'.format(id, seq))

def print_fastq_line(f, id, seq, qual):
    f.write('@{}\n{}\n+\n{}\n'.format(id, seq, qual))    
    
def sort_dict(mut_profiles):
    sorting_key = lambda item: 1000*{int:0, str:0, float:0, list:1, dict:2}[type(item[1])] + ord(item[0][0])
    mut_profiles = {k:mut_profiles[k] for k in sorted(mut_profiles)}
    mut_profiles = dict(sorted(mut_profiles.items(), key=sorting_key))
    for k, v in mut_profiles.items():
        if type(v) is not dict:
            continue
        mut_profiles[k] = sort_dict(v)
    return mut_profiles

def flatten_list_of_lists(v):
    return [item for sublist in v for item in sublist]

def empty_list_of_lists(L):
    return [[] for _ in range(L)]

def vector_of_zeros(L):
    return [0]*L

def count_pop_avg(mutations, sequence, section_start, section_end, base='ACGT'):
    # turn a list of lists into a list
    if base == 'N':
        base = 'ACGT'
    mutations = flatten_list_of_lists(mutations)
    out = [0]*len(sequence)

    for i in mutations:
        if next_base(sequence[i]) in base and section_start <= i < section_end:
            out[i] += 1
    return out
        
def count_mutations_per_read(mutations, section_start, section_end):
    # turn a list of lists into a list
    return [len([m for m in m_ if section_start <= m < section_end]) for m_ in mutations]


class Test_files_generator():
    
    def __init__(self, 
                 sample = 'my_test_sample',
                 path = os.getcwd(),
                 rnastructure_path = ''
                 ) -> None:
        
        self.sample = sample
        self.path = path
        self.rna = RNAstructure(rnastructure_path)

        # Sanity checks
        self.remove_files()
        os.makedirs(self.path, exist_ok=True)
        self.reference_count = 0
        self.sequences_stack = []
        self.mutations_stack = []
        self.reference_stack = []
        self.min_hamming_distance = 10
        
        self.samples_csv_file = os.path.join(self.path, 'samples.csv')
        self.library_file = os.path.join(self.path, 'library.csv')
        self.json_file = os.path.join(self.path, '{}.json'.format(self.sample))
        

    def iterate_reference_count(self):
        self.reference_count += 1
        return self.reference_count
        
    def remove_files(self):
        # assert that there's only json and fastq files in the folder
        if os.path.exists(self.path):
            for f in os.listdir(self.path):
                assert f.endswith('.json') or f.endswith('.fastq') or f.endswith('fasta') or f.endswith('.csv') or f == 'temp', "There are files in the folder that are not json or fastq files. Please remove them before running the test"
            os.system(' '.join(['rm', '-fr', self.path]))

    def create_read(self, mutations):
        read = self.sequence
        for m in mutations:
            read = read[:m] + next_base(read[m]) + read[m+1:]
        return read

    def assert_mutations_is_a_list_of_tuples(self, mutations):
        assert isinstance(mutations, list)
        for m in mutations:
            assert isinstance(m, tuple) or isinstance(m, list)
            for mm in m:
                assert isinstance(mm, int)
                assert mm < len(self.sequence)
                assert mm >= 0
            
    def generate_fastq_files(self, mutations):
        """Write a fastq file with the given parameters
        
        Arguments:
            self.path {str} -- where to write the fastq files
            self.mutations {dict} -- dictionary of mutations to be written in the fastq file
            
        """
        
        # Name the fastq files
        prefix = os.path.join(self.path, self.sample)
        fastq1_name, fastq2_name = prefix + '_R1.fastq', prefix + '_R2.fastq'

        # Write the fastq files
        with open(fastq1_name, 'a') as f1, open(fastq2_name, 'a') as f2:
            for i, m in enumerate(mutations):
                read = self.create_read(m)
                print_fastq_line(f1, '{}_read_{}'.format(self.reference_stack[-1], 1+i), read, 'F'*len(read))
                print_fastq_line(f2, '{}_read_{}'.format(self.reference_stack[-1], 1+i), invert_sequence(read), 'F'*len(read))
                
    def generate_fasta_files(self):
        """Write a fasta file with the given parameters
        
        Arguments:
            self.path {str} -- where to write the fasta files
            
        """
        
        # Name the fasta files
        file = os.path.join(self.path, self.sample + '.fasta')

        # Write the fasta files
        with open(file, 'a') as f:
            print_fasta_line(f, self.reference_stack[-1], self.sequence)
    
    def generate_samples_file(self):
        pd.DataFrame({
            'sample': [self.sample],
            'user': ['Napoleon'],
            'date': ['1769-08-15'],
            'exp_env': ['in_vitro'],
            'temperature_k': [310],
            'buffer': ['MgCl2'],
            'DMS_conc_mM': [2],
            'inc_time_tot_secs': [100]
        }, index=[0]).to_csv(self.samples_csv_file, index=False)
        
    def generate_library_file(self):
        df = {}
        for reference, sequence in zip(self.reference_stack, self.sequences_stack):
            df[reference] = {}
            df[reference]['reference'] = reference
            df[reference]['section_start'] = 1
            df[reference]['section_end'] = len(sequence)
            df[reference]['section'] = "{}-{}".format(1, len(sequence))
            df[reference]['some_random_attribute'] = 'some_random_value_{}'.format(reference)
        df = pd.DataFrame(df).T
        df.to_csv(self.library_file, index=False)
        
                
    def generate_json_file(self):
        
        if not os.path.isfile(self.samples_csv_file):
            self.generate_samples_file()
        samples_csv = pd.read_csv(self.samples_csv_file)
        samples_csv['temperature_k'] = samples_csv['temperature_k'].astype(int)
        
        if not os.path.isfile(self.library_file):
            self.generate_library_file()
        library = pd.read_csv(self.library_file)
        
        out = samples_csv.to_dict('records')[0]
        for idx, (reference, sequence, mutations) in enumerate(zip(self.reference_stack, self.sequences_stack, self.mutations_stack)):
            print(reference, sequence, mutations)
            n_reads = len(mutations)
            out[reference] = {}
            out[reference]['num_reads'] = n_reads
            out[reference]['num_aligned'] = n_reads
            out[reference]['skips_short_reads'] = 0
            out[reference]['skips_too_many_muts'] = 0
            out[reference]['skips_low_mapq'] = 0
            out[reference]['some_random_attribute'] = library['some_random_attribute'].values[idx]
            out[reference]['sequence'] = sequence   
            section = library['section'].values[idx]
            section_start = library['section_start'].values[idx]
            section_end = library['section_end'].values[idx]
            out[reference][section] = {}
            out[reference][section]['section_start'] = int(section_start)
            out[reference][section]['section_end'] = int(section_end)
            out[reference][section]['sequence'] = sequence[section_start-1:section_end]
            out[reference][section]['pop_avg'] = {}
            for base in ['A', 'C', 'G', 'T','N']:
                out[reference][section]['pop_avg']['sub_{}'.format(base)] = count_pop_avg(mutations, sequence, section_start-1, section_end, base)
            out[reference][section]['pop_avg']['sub_N'] = count_pop_avg(mutations, sequence, section_start-1, section_end, 'N')
            out[reference][section]['pop_avg']['sub_hist'] = np.histogram(count_mutations_per_read(mutations, section_start-1, section_end), bins=range(0, len(out[reference][section]['sequence'])))[0].tolist()
            out[reference][section]['pop_avg']['del'] = [0]*(section_end-section_start+1)
            out[reference][section]['pop_avg']['ins'] = [0]*(section_end-section_start+1)
            out[reference][section]['pop_avg']['cov'] = [n_reads]*(section_end-section_start+1)
            out[reference][section]['pop_avg']['info'] = [n_reads]*(section_end-section_start+1)
            out[reference][section]['pop_avg']['sub_rate'] = np.array( np.array(out[reference][section]['pop_avg']['sub_N'])/np.array(out[reference][section]['pop_avg']['info'])).tolist()
            out[reference][section]['pop_avg']['min_cov'] = min(out[reference][section]['pop_avg']['cov'])
            # RNAstructure
            rna_structure_prediction = self.rna.run(out[reference][section]['sequence'])
            out[reference][section]['structure'] = rna_structure_prediction['structure']
            print(rna_structure_prediction)
            print(out[reference][section]['sequence'])
            out[reference][section]['deltaG'] = float(rna_structure_prediction['deltaG'])
            out = sort_dict(out)
        with open(self.json_file, 'w') as f:
            json.dump(out, f, indent=2)
    
    def safe_sequence_generation(self, L):
        self.sequence = create_sequence(L)
        for past_sequence in self.sequences_stack:
            if len(past_sequence) != len(self.sequence):
                continue
            if hamming_distance(self.sequence, past_sequence) < self.min_hamming_distance:
                self.safe_sequence_generation(L)
        self.stack_sequence(self.sequence)
        return self.sequence
    
    def stack_sequence(self, sequence):
        self.sequences_stack.append(sequence)
        
    def stack_mutations(self, mutations):
        self.mutations_stack.append(mutations)
            
    def stack_reference(self):
        count = self.iterate_reference_count()
        self.reference_stack.append('reference_{}'.format(count))
            
    def add_reference(self, substitutions, L = 100):
        # Check and save inputs
        self.sequence = self.safe_sequence_generation(L)
        self.stack_mutations(substitutions)
        self.stack_reference()
        self.assert_mutations_is_a_list_of_tuples(substitutions)
        # Generate files
        self.generate_fastq_files(substitutions)
        self.generate_fasta_files()
        
if __name__ == '__main__':
    
    print('Generating test files to {}'.format(test_file_folder))
    
    t = Test_files_generator(
        path=test_file_folder, 
        sample='my_test_sample',
        rnastructure_path='/Users/ymdt/src/RNAstructure/exe',
        )
    t.generate_samples_file()
    t.add_reference(substitutions = [[4]]+[[9]]+[[14]]+[[19]]*4+[[24]]+[[29]]+[[34, 39]], L=50)
    t.add_reference(substitutions = [[3]]+[[8]]+[[13]]+[[18]]*4+[[23]]+[[28]]+[[33, 38]], L=50) # KEEP A SINGLE VALUE FOR L
    t.generate_library_file()
    t.generate_json_file()

    # Plot the data 
    s = Study(data = json.load(open(t.json_file, 'r')))
    for sa, re, se in zip(s.df['sample'].values, s.df['reference'].values, s.df['section'].values):
        s.mutation_fraction(sample=sa, reference=re, section=se)['fig'].show()



from hashlib import md5
import logging
import os
from typing import Any

import numpy as np
import pandas as pd


B_MATCH = 0
B_DELET = 1
B_INS_5 = 2
B_INS_3 = 3
B_SUB_A = 4
B_SUB_C = 5
B_SUB_G = 6
B_SUB_T = 7


def get_files(folder: str, ext: str):
    paths = []
    for reference in os.listdir(folder):
        if os.path.isdir(os.path.join(folder, reference)):
            for section in os.listdir(os.path.join(folder, reference)):
                if not section.endswith(ext):
                    continue
                paths.append(os.path.join(folder, reference, section))
    else:
        if reference.endswith(ext):
            paths.append(os.path.join(folder, reference))
    return paths


def fastq_to_df(fastq_file):    
    df, data = pd.DataFrame(), pd.read_csv(fastq_file, sep='\t', header=None)
    df['reference'] = data.iloc[np.arange(0,len(data),4)].reset_index(drop=True)
    df['sequence'] = data.iloc[np.arange(1,len(data),4)].reset_index(drop=True)
    df['quality'] = data.iloc[np.arange(3,len(data),4)].reset_index(drop=True)
    return df


def sam_to_df(path):
    skiprows = 0
    with open(path) as f:
        while f.readline().startswith('@'):
            skiprows += 1
    df = pd.read_csv(path, sep='\t', skiprows=skiprows, header=None)
    df = df[df.columns[:11]]
    df.columns = ['QNAME', 'FLAG', 'RNAME', 'POS', 'MAPQ', 'CIGAR', 'RNEXT', 'PNEXT','TLEN', 'SEQ', 'QUAL']
    return df


def sort_fastq_pairs(fq1s, fq2s):
    if type(fq1s) is str:
        fq1s = [fq1s]
    if type(fq2s) is str:
        fq2s = [fq2s]
    assert len(fq1s) >= len(fq2s), 'More fq2s than fq1s'
    for f2 in fq2s:
        if f2.replace('_R2', '_R1') not in fq1s:
            raise ValueError(f'No matching pair for {f2}')
    fq1s = [f2.replace('_R2', '_R1') for f2 in fq2s] + [f for f in fq1s if f.replace('_R1','_R2') not in fq2s]
    fq2s += [None for f in fq1s if f.replace('_R1','_R2') not in fq2s]
    samples = [f.split('/')[-1].split('.')[0].split('_')[:-1] for f in fq1s]
    samples = ['_'.join(s) for s in samples]
    return fq1s, fq2s, samples


def query_muts(muts: np.ndarray, bits: int, sum_up = True, axis=0, set_type = 'superset'):
    """
    Count the number of times a query mutation occurs in each column
    or one column of a set of mutation mut_vectors.
    The counting operation comprises three steps:
    1. bitwise AND to confirm at least one "1" bit is shared, e.g.
       bits: 11110000 & mut_vectors: 00100000 -> 00100000 (True)
       bits: 11110000 & mut_vectors: 00100010 -> 00100000 (True)
       bits: 11110000 & mut_vectors: 00000000 -> 00000000 (False)
    2. bitwise OR to confirm no "1" bit in mut_vectors is not in bits, e.g.
       bits: 11110000 | mut_vectors: 00100000 -> 11110000 =? 11110000 (True)
       bits: 11110000 | mut_vectors: 00100010 -> 11110010 =? 11110000 (False)
       bits: 11110000 | mut_vectors: 00000000 -> 11110000 =? 11110000 (True)
    3. logical AND to confirm that both tests pass, e.g.
       bits: 11110000, mut_vectors: 00100000 -> True  AND True  (True)
       bits: 11110000, mut_vectors: 00100010 -> True  AND False (False)
       bits: 11110000, mut_vectors: 00000000 -> False AND True  (False)

    Arguments
    mut_vectors: NDArray of a set of mutation mut_vectors (2-dimensional)
          or one column in a set of mutation mut_vectors (1-dimensional).
          Data type must be uint8.
    bits: One-byte int in the range [0, 256) representing the mutation
          to be queried. The bits in the int encode the mutation as
          defined above, e.g.
          - 00000010 (int 2) is a deletion
          - 11010001 (int 209) is either substitution to A, G, or T
                               or a match to C
    
    Returns
    if sum_up: 
        int of the number of times the query mutation occurs in mut_vectors
        count: If mut_vectors is 1-dimensional, int of the number of times the
            query mutation occurs in mut_vectors.
            If mut_vectors is 2-dimensional, NDArray with one int for each
            column in mut_vectors.
    if not sum_up:
        bool NDArray with one bool for each column in mut_vectors.
        True if the query mutation occurs in the column, False otherwise.
    """
    if not muts.dtype == np.uint8:
        raise TypeError('mut_vectors must be of type uint8 and not {}'.format(muts.dtype))
    #print(np.logical_and(mut_vectors & bits, (mut_vectors | bits) == bits).sum(axis=0))
    assert isinstance(bits, int) and 0 <= bits < 256

        
    if set_type == 'subset':
        logic_fun = lambda m, b: np.logical_and(m & b, (m | b) == b)
    if set_type == 'superset':
        logic_fun = lambda m, b: np.array(m & b, dtype=bool)
    
    if sum_up:
        return logic_fun(muts, bits).sum(axis=axis)
    else:
        return logic_fun(muts, bits)


def get_num_parallel(n_tasks: int,
                     max_procs: int,
                     parallel: bool,
                     hybrid: bool = False) -> tuple[int, int]:
    """ Determine how to parallelize the tasks.

    Parameters
    ----------
    n_tasks: int (≥ 1)
        Number of tasks to parallelize
    max_procs: int (≥ 1)
        Maximum number of processes to run at one time
    parallel: bool
        Whether to permit multiple tasks to be run in parallel
    hybrid: bool (default: False)
        Whether to allow both multiple tasks to run in parallel and,
        at the same, each task to run multiple processes in parallel

    Returns
    -------
    int (≥ 1)
        Number of tasks to run in parallel
    int (≥ 1)
        Number of processes to run for each task
    """
    if n_tasks >= 1 and max_procs >= 1:
        # This function only works if there is at least one task to
        # parallelize, at least one process is allowed, and parallel
        # is a valid option.
        if parallel:
            # Multiple tasks may be run in parallel. The number of tasks
            # run in parallel cannot exceed 1) the total number of tasks
            # and 2) the user-specified maximum number of processes.
            n_tasks_parallel = min(n_tasks, max_procs)
        else:
            # Otherwise, only one task at a time can be run.
            n_tasks_parallel = 1
        if n_tasks_parallel == 1 or hybrid:
            # Each individual task can be run by multiple processes in
            # parallel, as long as either 1) multiple tasks are not run
            # simultaneously in parallel (i.e. n_tasks_parallel == 1)
            # or 2) the calling function sets hybrid=True, which lets
            # multiple tasks run in parallel and each run with multiple
            # processes. Only the alignment module can simultaneously
            # run multiple tasks and multiple processes for each task
            # because its two most computation-heavy processes (cutadapt
            # and bowtie2) come with their own parallelization abilities
            # that can work independently of Python's multiprocessing
            # module. However, the other modules (e.g. vectoring) are
            # parallelized using the multiprocessing module, which does
            # not support "nesting" parallelization in multiple layers.
            # Because n_tasks_parallel is either 1 or the smaller of
            # n_tasks and n_procs (both of which are ≥ 1), it must be
            # that 1 ≤ n_tasks_parallel ≤ n_procs, and therefore that
            # 1 ≤ n_procs / n_tasks_parallel ≤ n_procs, so the
            # integer quotient must be a valid number of processes.
            n_procs_per_task = max_procs // n_tasks_parallel
        else:
            # Otherwise, only one process can work on each task.
            n_procs_per_task = 1
    else:
        logging.warning("Defaulting to 1 process due to invalid number of "
                        f"tasks ({n_tasks}) and/or processes ({max_procs}).")
        n_tasks_parallel = 1
        n_procs_per_task = 1
    return n_tasks_parallel, n_procs_per_task


def digest_file(file_path: Any) -> str:
    """
    Compute the checksum of a file.

    Parameters
    ----------
    file_path: Any (path-like)
        Path of the file on which to compute the checksum. Can be
        any type that the open() function recognizes as a path.

    Returns
    -------
    str
        Checksum of the file (in hexadecimal)
    """
    with open(file_path, "rb") as f:
        digest = md5(f.read()).hexdigest()
    return digest



import click

from ..util.cli import *
from ..stats.main import run


@click.command()
@opt_out_dir
@opt_stats_count
@opt_stats_frac
@opt_report
def cli(*args, **kwargs):
    run(*args, **kwargs)


if __name__ == "__main__":
    cli()



from pathlib import Path
import sys

from ..util.path import BasePath


# Byte encodings for nucleic acid alphabets
BASES = b"ACGT"
COMPS = b"TGCA"
RBASE = b"ACGU"
RCOMP = b"UGCA"
BASEN = b"N"
BASES_SET = set(BASES)
BASEN_SET = set(BASES + BASEN)

# Integer encodings for nucleic acid alphabets
A_INT = BASES[0]
C_INT = BASES[1]
G_INT = BASES[2]
T_INT = BASES[3]
N_INT = BASEN[0]

# Byte encodings for mutation mut_vectors
BLANK = b"\x00"  # 00000000 (000): no coverage at this position
MATCH = b"\x01"  # 00000001 (001): match with reference
DELET = b"\x02"  # 00000010 (002): deletion from reference
INS_5 = b"\x04"  # 00000100 (004): insertion 5' of base in reference
INS_3 = b"\x08"  # 00001000 (008): insertion 3' of base in reference
SUB_A = b"\x10"  # 00010000 (016): substitution to A
SUB_C = b"\x20"  # 00100000 (032): substitution to C
SUB_G = b"\x40"  # 01000000 (064): substitution to G
SUB_T = b"\x80"  # 10000000 (128): substitution to T

# Integer encodings for mutation mut_vectors
BLANK_INT = BLANK[0]
MATCH_INT = MATCH[0]
DELET_INT = DELET[0]
INS_5_INT = INS_5[0]
INS_3_INT = INS_3[0]
SUB_A_INT = SUB_A[0]
SUB_C_INT = SUB_C[0]
SUB_G_INT = SUB_G[0]
SUB_T_INT = SUB_T[0]

# Ambiguous encodings for mutation mut_vectors
SUB_N_INT = SUB_A_INT | SUB_C_INT | SUB_G_INT | SUB_T_INT
SUB_N = SUB_N_INT.to_bytes(length=1, byteorder=sys.byteorder)
ANY_N_INT = MATCH_INT | SUB_N_INT
ANY_N = ANY_N_INT.to_bytes(length=1, byteorder=sys.byteorder)
AMBIG_INT = ANY_N_INT | DELET_INT | INS_5_INT | INS_3_INT
AMBIG = AMBIG_INT.to_bytes(length=1, byteorder=sys.byteorder)


def get_diffs(seq1, seq2):
    if len(seq1) != len(seq2):
        raise ValueError("Sequences were different lengths: "
                         f"'{seq1}' and '{seq2}'")
    diffs = [i for i, (x1, x2) in enumerate(zip(seq1, seq2)) if x1 != x2]
    return diffs


class Seq(bytes):
    __slots__ = []

    alph = b""
    comp = b""
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def __init__(self, seq: bytes):
        self.validate_seq(seq)
        super().__init__()
    
    @classmethod
    def validate_seq(cls, seq):
        if not seq:
            raise ValueError("seq is empty")
        if set(seq) - cls.alphaset:
            raise ValueError(f"Invalid characters in seq: '{seq.decode()}'")

    @property
    def rc(self):
        return self.__class__(self[::-1].translate(self.trans))

    def __getitem__(self, item):
        return self.__class__(super().__getitem__(item))

    def __str__(self):
        return self.decode()


class DNA(Seq):
    alph = BASES
    comp = COMPS
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def tr(self):
        """ Transcribe DNA to RNA. """
        return RNA(self.replace(b"T", b"U"))


class RNA(Seq):
    alph = RBASE
    comp = RCOMP
    alphaset = set(alph)
    trans = alph.maketrans(alph, comp)

    def rt(self):
        """ Reverse transcribe RNA to DNA. """
        return DNA(self.replace(b"U", b"T"))


class FastaIO(object):
    recsym = b">"
    deftrunc = len(recsym)

    def __init__(self, path: str | Path | BasePath):
        if path is None:
            # If the user forgets to give a FASTA file when one is
            # required, then the program will crash with an ugly error
            # when it tries to open a path that is None. This check
            # raises an error that describes the specific problem.
            raise TypeError("No FASTA file was given.")
        self._path = path


class FastaParser(FastaIO):
    def __init__(self, path: str | Path | BasePath):
        super().__init__(path)
        self._refs: set[str] = set()

    @classmethod
    def _parse_fasta_record(cls, fasta, line: bytes):
        if not line.startswith(cls.recsym):
            raise ValueError("FASTA definition line does not start with "
                             f"'{cls.recsym.decode()}'")
        name = line.split()[0][cls.deftrunc:].decode()
        seq = bytearray()
        while (line := fasta.readline()) and not line.startswith(cls.recsym):
            seq.extend(line.rstrip())
        seq = DNA(bytes(seq))
        return line, name, seq

    def parse(self):
        with open(self._path, "rb") as f:
            line = f.readline()
            while line:
                line, name, seq = self._parse_fasta_record(f, line)
                if name in self._refs:
                    raise ValueError(
                        f"Duplicate entry in {self._path}: '{name}'")
                self._refs.add(name)
                yield name, seq


class FastaWriter(FastaIO):
    def __init__(self, path: str, refs: dict[str, DNA]):
        super().__init__(path)
        self._refs = refs
    
    def write(self):
        with open(self._path, "wb") as f:
            for ref, seq in self._refs.items():
                f.write(b"".join(
                    [self.recsym, ref.encode(), b"\n", seq, b"\n"]
                ))
                

def parse_fasta(path: str):
    return FastaParser(path).parse()



import logging
import shlex
import subprocess

from typing import Any, List


# Commands for external applications

BOWTIE2_CMD = "bowtie2"
BOWTIE2_BUILD_CMD = "bowtie2-build"
CUTADAPT_CMD = "cutadapt"
FASTQC_CMD = "fastqc"
SAMTOOLS_CMD = "samtools"


# Command utility functions

def run_cmd(args: List[Any], shell: bool = False):
    args_str = tuple(map(str, args))
    cmd_str = shlex.join(args_str)
    logging.debug(f"Running command (shell={shell}):\n{cmd_str}")
    subprocess.run(args_str, check=True, shell=shell)
    return cmd_str



from collections import defaultdict
from functools import wraps
from inspect import getmembers, Parameter, Signature
import logging
from textwrap import dedent
from typing import Any, Callable, Iterable

from click import Option

from ..util import cli

# Ignore special parameters with reserved names.
reserved_params = {"self", "cls"}

# Get every option defined for the command line interface.
cli_options = dict(getmembers(cli, lambda member: isinstance(member, Option)))

# Get the default value for every parameter.
api_defs = {
    "n_procs": cli.NUM_CPUS,
}
cli_defs = {option.name: option.default for option in cli_options.values()
            if option.default is not None}
all_defs = {**cli_defs, **api_defs}

# Get the documentation for every parameter.
cli_docs = {option.name: option.help for option in cli_options.values()}
api_docs = {
    "bytes_per_batch": "Number of bytes per batch of mutation vectors",
    "fq_unit": "FASTQ file or pair of mated FASTQ files",
    "fq_units": "Iterable of FASTQ files or pairs of mated FASTQ files",
    "kwargs": "Additional keyword-only arguments",
    "n_procs": "Number of processes on which to run",
}
all_docs = {**cli_docs, **api_docs}


def get_param_default(param: Parameter,
                      defaults: dict[str, Any],
                      exclude_defs: tuple[str, ...]) -> Parameter:
    """ Return the parameter, possibly with a new default value. """
    if param.name in exclude_defs:
        logging.debug(f"Skipped excluded parameter '{param.name}'")
        return param
    if param.name in reserved_params:
        logging.debug(f"Skipped reserved parameter '{param.name}'")
        return param
    try:
        default = defaults[param.name]
    except KeyError:
        logging.debug(f"Skipped defaultless parameter '{param.name}'")
        return param
    logging.debug(f"Set parameter '{param.name}' default to {repr(default)}")
    # Return copy of parameter with new default value.
    return param.replace(default=default)


def paramdef(defaults: dict[str, Any], exclude_defs: tuple[str, ...]):
    """ Give the keyword argments of a function default values. """
    if defaults is None:
        defaults = dict()

    def decorator(func: Callable):
        logging.debug("Setting default values of parameters for function "
                      f"'{func.__name__}'")
        # List all the parameters of the function, replacing the default
        # value of those parameters with defaults given in defaults.
        sig = Signature.from_callable(func)
        new_params = [get_param_default(param, defaults, exclude_defs)
                      for param in sig.parameters.values()]

        # Update the help text (does not affect actual default values).
        try:
            func.__signature__ = Signature(parameters=new_params)
        except ValueError as error:
            raise ValueError(f"Failed to set signature of {func.__name__} to "
                             f"({', '.join(map(str, new_params))}): {error}")

        # Update the actual default values of keyword-only arguments
        # (does not affect help text).
        default_kwargs = {param.name: param.default for param in new_params
                          if param.kind == Parameter.KEYWORD_ONLY}

        @wraps(func)
        def new_func(*args, **kwargs):
            return func(*args, **{**default_kwargs, **kwargs})

        return new_func

    return decorator


def autodef(extra_defs: dict[str, Any] | None = None,
            exclude_defs: tuple[str, ...] = ()):
    """ Call ```paramdef``` and automatically infer default values from
    the CLI and API. Extra defaults (if needed) may be given as keyword
    arguments. """
    return paramdef(all_defs if extra_defs is None
                    else {**all_defs, **extra_defs},
                    exclude_defs)


def get_param_lines(func: Callable, param_docs: dict[str, str]):
    sig = Signature.from_callable(func)
    # Add information about every parameter to the docstring.
    param_lines = list()
    for name, param in sig.parameters.items():
        if name in reserved_params:
            # Ignore reserved parameters (if any).
            continue
        if doc := param_docs.get(name):
            # Add the type annotation (if any) after the name of the
            # parameter.
            if param.annotation is param.empty:
                name_type = name
            else:
                try:
                    name_type = f"{name}: {param.annotation.__name__}"
                except AttributeError:
                    # Some types (e.g. UnionType) have no name.
                    name_type = f"{name}: {param.annotation}"
            # Add the default value (if any) in brackets after the
            # documentation of the parameter.
            if param.default is not param.empty:
                doc = f"{doc}  [default: {repr(param.default)}]"
            # Add the parameter's name, type, kind, and documentation to
            # the docstring.
            param_lines.extend([f"{name_type}  [{param.kind.description}]",
                                f"    {doc}"])
        else:
            logging.warning("Missing documentation for parameter "
                            f"'{name}' of function '{func.__name__}'")
    return param_lines


def get_doc_lines(func: Callable, param_lines: list[str], return_doc: str):
    sig = Signature.from_callable(func)
    doc_lines = list()
    if func.__doc__:
        # Use the existing docstring to start the new docstring.
        doc_lines.append(dedent(func.__doc__))
    else:
        logging.warning(f"Function '{func.__name__}' had no docstring")
    if param_lines:
        if doc_lines:
            doc_lines.append("")
        doc_lines.extend(["Parameters",
                          "----------"])
        doc_lines.extend(param_lines)
    if sig.return_annotation is not sig.empty:
        if return_doc:
            if doc_lines:
                doc_lines.append("")
            doc_lines.extend(["Return",
                              "------",
                              f"{sig.return_annotation}",
                              f"    {return_doc}"])
        else:
            logging.warning(
                f"Function '{func.__name__}' has return annotation "
                f"{sig.return_annotation} but was given no return text")
    elif return_doc:
        logging.warning(f"Function '{func.__name__}' was given return text "
                        f"'{return_doc}' but has no return annotation")
    return doc_lines


def paramdoc(param_docs: dict[str, str], return_doc: str):
    """
    Give a function a new docstring where each parameter gets annotated
    with the text given in the keyword arguments (```param_docs```).

    Parameters
    ----------
    param_docs: dict[str, str]
        Description of each parameter, keyed by name
    return_doc: str
        Description of the return value; will occur at end of docstring

    Return
    ------
    callable
        Function with new docstring
    """

    def decorator(func: Callable):
        param_lines = get_param_lines(func, param_docs)
        doc_lines = get_doc_lines(func, param_lines, return_doc)
        func.__doc__ = "\n".join(doc_lines)
        return func

    return decorator


def autodoc(extra_docs: dict[str, str] | None = None, return_doc: str = ""):
    """ Call ```paramdoc``` and automatically infer descriptions and
    type annotations about all parameters from the CLI and API.
    Documentation of any extra parameters may also be given. """
    return paramdoc(all_docs if extra_docs is None
                    else {**all_docs, **extra_docs},
                    return_doc)


def auto(*,
         extra_defs: dict[str, Any] | None = None,
         exclude_defs: tuple[str, ...] = (),
         extra_docs: dict[str, str] | None = None,
         return_doc: str = ""):
    """ Combine ```autodef``` and ```autodoc```, in that order. """

    def decorator(func: Callable):
        func = autodef(extra_defs, exclude_defs)(func)
        func = autodoc(extra_docs, return_doc)(func)
        return func

    return decorator



from enum import Enum
import itertools
import logging
import re
from functools import cached_property
from typing import BinaryIO, Dict, Iterable

from ..util import cli, docdef, path
from ..util.excmd import (run_cmd, BOWTIE2_CMD, BOWTIE2_BUILD_CMD,
                          CUTADAPT_CMD, FASTQC_CMD, SAMTOOLS_CMD)
from ..util.seq import FastaParser

# SAM file format specifications
SAM_HEADER = b"@"
SAM_ALIGN_SCORE = b"AS:i:"
SAM_EXTRA_SCORE = b"XS:i:"

# Bowtie2 parameters
MATCH_BONUS = "1"
MISMATCH_PENALTY = "1,1"
N_PENALTY = "0"
REF_GAP_PENALTY = "0,1"
READ_GAP_PENALTY = "0,1"


def index_bam_file(bam: path.BasePath):
    bam_index = bam.edit(ext=path.BAI_EXT)
    cmd = [SAMTOOLS_CMD, "index", bam]
    run_cmd(cmd)
    if not bam_index.path.is_file():
        logging.critical(f"Failed to index file: {bam}")


class FastqUnit(object):
    """
    Unified interface for handling the following sets of sequencing reads:

    - One FASTQ file of single-end reads from one sample
    - One FASTQ file of interleaved, paired-end reads from one sample
    - Two FASTQ files of mate 1 and mate 2 paired-end reads from one sample
    - One FASTQ file of single-end reads originating from one reference sequence
      in one sample
    - One FASTQ file of interleaved, paired-end reads originating from one
      reference sequence in one sample
    - Two FASTQ files of mate 1 and mate 2 paired-end reads originating from one
      reference sequence in one sample

    """

    MAX_PHRED_ENC = 127  # 2^7 - 1

    class KeyName(Enum):
        SINGLE = "fastqs"  # single-end reads
        INTER = "fastqi"  # interleaved paired-end reads
        MATE1 = "fastq1"  # mate 1 paired-end reads
        MATE2 = "fastq2"  # mate 2 paired-end reads

    class Bowtie2Flag(Enum):
        SINGLE = "-U"
        INTER = "--interleaved"
        MATE1 = "-1"
        MATE2 = "-2"

    class KeySampleType(Enum):
        SINGLE = path.AbstractSampleReadsFilePath
        INTER = path.AbstractSampleReadsFilePath
        MATE1 = path.AbstractSampleReads1FilePath
        MATE2 = path.AbstractSampleReads2FilePath

    class KeySampleInType(Enum):
        SINGLE = path.SampleReadsInFilePath
        INTER = path.SampleReadsInFilePath
        MATE1 = path.SampleReads1InFilePath
        MATE2 = path.SampleReads2InFilePath

    class KeyOneRefType(Enum):
        SINGLE = path.AbstractOneRefReadsFilePath
        INTER = path.AbstractOneRefReadsFilePath
        MATE1 = path.AbstractOneRefReads1FilePath
        MATE2 = path.AbstractOneRefReads2FilePath

    class KeySampleDirName(Enum):
        SINGLE = "fastqs_dir"  # single-end reads
        INTER = "fastqi_dir"  # interleaved paired-end reads
        MATE12 = "fastq12_dir"  # mate 1 and mate 2 paired-end reads

    class KeySampleDirType(Enum):
        SINGLE = path.AbstractSampleDirPath
        INTER = path.AbstractSampleDirPath
        MATE12 = path.AbstractSampleDirPath

    def __init__(self, /, *,
                 fastqs: (path.AbstractSampleReadsFilePath |
                          path.AbstractOneRefReadsFilePath |
                          None) = None,
                 fastqi: (path.AbstractSampleReadsFilePath |
                          path.AbstractOneRefReadsFilePath |
                          None) = None,
                 fastq1: (path.AbstractSampleReads1FilePath |
                          path.AbstractOneRefReads1FilePath |
                          None) = None,
                 fastq2: (path.AbstractSampleReads2FilePath |
                          path.AbstractOneRefReads2FilePath |
                          None) = None,
                 phred_enc: int):
        self._phred_enc = phred_enc
        self._init_inputs = dict(zip(self._get_keyname_strs(),
                                     (fastqs, fastqi, fastq1, fastq2),
                                     strict=True))
        self._validate_inputs()

    def _validate_inputs(self):
        _ = self.phred_arg
        _ = self.inputs
        _ = self.sample
        _ = self.ref

    def dict(self):
        return dict(sample=self.sample, ref=self.ref)

    @property
    def phred_enc(self):
        if 0 <= self._phred_enc <= self.MAX_PHRED_ENC:
            return self._phred_enc
        raise ValueError(self._phred_enc)

    @property
    def phred_arg(self):
        return f"--phred{self.phred_enc}"

    @classmethod
    def _get_keyname_strs(cls):
        return tuple(str(key.value) for key in cls.KeyName)

    @classmethod
    def _test_single(cls, keys: tuple[str, ...]):
        """ Return whether the arguments match a single-end FASTQ.
        No validation is performed. """
        return keys == (cls.KeyName.SINGLE.value,)

    @classmethod
    def _test_interleaved(cls, keys: tuple[str, ...]):
        """ Return whether the arguments match an interleaved FASTQ.
        No validation is performed. """
        return keys == (cls.KeyName.INTER.value,)

    @classmethod
    def _test_separate_mates(cls, keys: tuple[str, ...]):
        """ Return whether the arguments match two separate paired-end
        FASTQs. No validation is performed. """
        return keys == (cls.KeyName.MATE1.value, cls.KeyName.MATE2.value)

    @cached_property
    def _input_keys(self):
        """ Validate that a correct set of keyword arguments was given, and
        then return them. """
        # Find all the keys of self._inputs that got a value (not None).
        keys = tuple(key for key in self._get_keyname_strs()
                     if self._init_inputs[key] is not None)
        # The keys must match one of the following patterns to be valid.
        if not any((self._test_single(keys),
                    self._test_interleaved(keys),
                    self._test_separate_mates(keys))):
            raise ValueError(f"Got invalid FASTQ keywords: {keys}")
        # Now the keys are validated, and can validate other properties.
        return keys

    @cached_property
    def _inputs(self) -> Dict[str, (path.AbstractSampleReadsFilePath |
                                    path.AbstractOneRefReadsFilePath |
                                    path.AbstractSampleReads1FilePath |
                                    path.AbstractOneRefReads1FilePath |
                                    path.AbstractSampleReads2FilePath |
                                    path.AbstractOneRefReads2FilePath)]:
        """ Return a dictionary of keyword arguments and path instances of the
        input FASTQ files. Keys are validated, but values and types are not. """
        return {key: self._init_inputs[key] for key in self._input_keys}

    @property
    def single(self):
        """ Return whether the arguments match a single-end FASTQ.
        Validated. """
        return self._test_single(self._input_keys)

    @property
    def interleaved(self):
        """ Return whether the arguments match an interleaved FASTQ.
        Validated. """
        return self._test_interleaved(self._input_keys)

    @property
    def separate_mates(self):
        """ Return whether the arguments match two separate paired-end FASTQs.
        Validated. """
        return self._test_separate_mates(self._input_keys)

    @property
    def paired(self):
        """ Return whether the reads in the FASTQ unit are paired-end.
        Validated. """
        return not self.single

    @classmethod
    def _test_by_sample(cls, keys: tuple[str, ...], inputs: Iterable):
        """ Return whether the given path types match the path types expected
        for FASTQ files representing an entire sample. """
        return all(isinstance(i, cls.KeySampleType[cls.KeyName(key).name].value)
                   for key, i in zip(keys, inputs, strict=True))

    @classmethod
    def _test_by_ref(cls, keys: tuple[str, ...], inputs: Iterable):
        """ Return whether the given path types match the path types expected
        for FASTQ files representing one reference from a sample. """
        return all(isinstance(i, cls.KeyOneRefType[cls.KeyName(key).name].value)
                   for key, i in zip(keys, inputs, strict=True))

    @classmethod
    def _by_sample_or_ref(cls, keys: tuple[str, ...], inputs: Iterable):
        """ Validate that the types of paths given as input arguments
        match the expected types. """
        # The types must match one of the expected set of path types
        # given the keyword arguments -- either the sample- or ref-
        # based paths.
        by_sample = cls._test_by_sample(keys, inputs)
        by_ref = cls._test_by_ref(keys, inputs)
        if by_sample == by_ref:
            raise TypeError(f"Got invalid input FASTQ types: {inputs}")
        # Now the types are validated and can validate other properties.
        return by_sample, by_ref

    @cached_property
    def by_ref(self):
        _, by_ref = self._by_sample_or_ref(self._input_keys,
                                           tuple(self._inputs.values()))
        return by_ref

    @property
    def by_sample(self):
        return not self.by_ref

    @property
    def inputs(self):
        """ Return a dictionary of keyword arguments and path instances
        of the input FASTQ files. Validated. """
        # Confirm that the inputs either match the expected types for
        # sample- or ref-based inputs.
        _ = self.by_ref
        # Now the inputs are validated.
        return self._inputs

    @property
    def paths(self):
        return tuple(inp.path for inp in self.inputs.values())

    @property
    def parent(self) -> path.BasePath:
        parents = [inp.parent for inp in self.inputs.values()]
        if not parents:
            raise TypeError("Not parent directory")
        if any(parent != parents[0] for parent in parents[1:]):
            raise ValueError("More than one parent directory")
        return parents[0]

    @property
    def str_paths(self):
        return tuple(map(str, self.inputs.values()))

    @cached_property
    def sample(self):
        samples: tuple[str] = tuple({fq.sample for fq in self.inputs.values()})
        if len(samples) != 1:
            raise ValueError(f"Required exactly one sample, but got {samples}")
        return samples[0]

    @cached_property
    def ref(self):
        if not self.by_ref:
            return None
        refs: tuple[str] = tuple({fq.ref for fq in self.inputs.values()})
        if len(refs) != 1:
            raise ValueError(f"Required exactly one reference, but got {refs}")
        return refs[0]

    def compatible_fasta(self, fasta: (path.AbstractRefsetSeqFilePath |
                                       path.OneRefSeqStepFilePath)):
        """ Return whether a given FASTA file is compatible with the
        FastqUnit. They are compatible if and only if
        - the FASTA has a set of references and the FASTQ has reads from
          an entire sample
        - the FASTA has one reference and the FASTQ has reads from only
          the same reference
        """
        if isinstance(fasta, path.AbstractRefsetSeqFilePath):
            return self.by_sample
        if isinstance(fasta, path.OneRefSeqStepFilePath):
            return self.ref == fasta.ref
        raise TypeError(fasta)

    @property
    def cutadapt_input_args(self):
        return self.paths

    @property
    def _bowtie2_flags(self):
        return tuple(str(self.Bowtie2Flag[self.KeyName(key).name].value)
                     for key in self._input_keys)

    @property
    def bowtie2_inputs(self):
        return tuple(itertools.chain(*map(list, zip(self._bowtie2_flags,
                                                    self.paths,
                                                    strict=True))))

    def edit(self, **fields):
        """
        Return a new FastqUnit by modifying all the FASTQ paths, using
        the given fields.

        Parameters
        ----------
        **fields: Any
            Keyword arguments to determine the new paths

        Returns
        -------
        FastqUnit
            New FastqUnit with the same keys and modified paths
        """
        return FastqUnit(**{key: fq.edit(**{**fields, path.EXT_KEY: fq.ext})
                            for key, fq in self.inputs.items()},
                         phred_enc=self.phred_enc)

    @classmethod
    def _get_demult_files(cls, phred_enc: int, dir_path: path.BasePath,
                          key: str):
        """
        Yield a FastqUnit for each demultiplexed FASTQ file in a
        directory.

        Parameters
        ----------
        phred_enc: int
            Phred score encoding
        dir_path: SampleInDirPath
            Directory containing the FASTQ files
        key: str
            Keyword argument for every FASTQ file ('fastqs' or 'fastqi')

        Yield
        -----
        FastqUnit
            One for each FASTQ file in the directory
        """
        dp = dir_path.path
        parse_type = cls.KeyOneRefType[cls.KeyName(key).name].value
        for fname in dp.iterdir():
            yield cls(**{key: parse_type.parse(dp.joinpath(fname))},
                      phred_enc=phred_enc)

    @classmethod
    def _get_demult_pairs(cls, phred_enc: int, dir_path: path.BasePath):
        """
        Yield a FastqUnit for each pair of mated, demultiplexed FASTQ files
        in a directory.

        Parameters
        ----------
        phred_enc: int
            Phred score encoding
        dir_path: SampleInDirPath
            Directory containing the FASTQ files

        Return
        ------
        Iterable[FastqUnit]
            One for each FASTQ pair in the directory
        """
        dp = dir_path.path
        # Create empty dictionaries to store the FASTQ files for mates 1 and 2,
        # keyed by the name of the reference sequence for each FASTQ file.
        mates1: dict[str, path.AbstractOneRefReads1FilePath] = dict()
        mates2: dict[str, path.AbstractOneRefReads2FilePath] = dict()
        # Read the name of every file in the directory.
        for fname in dp.iterdir():
            # Get the full path to the file.
            fpath = str(dp.joinpath(fname))
            # Determine if the file is a mate 1 or mate 2 FASTQ file.
            is1 = any(fpath.endswith(ext) for ext in path.FQ1_EXTS)
            is2 = any(fpath.endswith(ext) for ext in path.FQ2_EXTS)
            if is1 and is2:
                # There should be no way for this error to happen, but catching
                # it just in case.
                raise ValueError(f"FASTQ path matched both mates: '{fpath}'")
            if is1:
                # The file name matched an extension for a mate 1 FASTQ file.
                fq = path.AbstractOneRefReads1FilePath.parse(fpath)
                if fq.ref in mates1:
                    raise ValueError(f"Got >1 FASTQ file for ref '{fq.ref}'")
                # Add the path to the dict of mate 1 files, keyed by reference.
                mates1[fq.ref] = fq
            elif is2:
                # The file name matched an extension for a mate 2 FASTQ file.
                fq = path.AbstractOneRefReads2FilePath.parse(fpath)
                if fq.ref in mates2:
                    raise ValueError(f"Got >1 FASTQ file for ref '{fq.ref}'")
                # Add the path to the dict of mate 2 files, keyed by reference.
                mates2[fq.ref] = fq
            else:
                # If a file name does not match the expected FASTQ name format,
                # log a warning but keep going, since the presence or absence
                # of one FASTQ file will not affect the others, and this file
                # might just be an extraneous file, such as .DS_Store on macOS.
                logging.warning(f"File name is not a valid FASTQ: '{fpath}'")
        # Iterate through all references from mate 1 and mate 2 files.
        for ref in set(mates1) | set(mates2):
            fastq1 = mates1.get(ref)
            fastq2 = mates2.get(ref)
            if fastq1 is not None and fastq2 is not None:
                # Yield a new FastqUnit from the paired FASTQ files.
                yield cls(fastq1=fastq1, fastq2=fastq2, phred_enc=phred_enc)
            elif fastq1 is None:
                logging.error(f"Missing mate 1 for reference '{ref}'")
            else:
                logging.error(f"Missing mate 2 for reference '{ref}'")

    @classmethod
    def _get_sample_files(cls, phred_enc: int, path_strs: tuple[str], key: str):
        file_type = cls.KeySampleInType[cls.KeyName(key).name].value
        for path_str in path_strs:
            yield FastqUnit(**{key: file_type.parse(path_str)},
                            phred_enc=phred_enc)

    @classmethod
    def _get_sample_pairs(cls, phred_enc: int,
                          path1_strs: tuple[str],
                          path2_strs: tuple[str]):
        keys = (cls.KeyName.MATE1, cls.KeyName.MATE2)
        for path_strs in zip(path1_strs, path2_strs, strict=True):
            paths = {
                key.value: cls.KeySampleInType[key.name].value.parse(path_str)
                for key, path_str in zip(keys, path_strs, strict=True)
            }
            yield FastqUnit(**paths, phred_enc=phred_enc)

    @classmethod
    def _from_strs(cls, /, *, phred_enc: int, **fastq_args: tuple[str]):
        """ Yield a FastqUnit for each FASTQ file (or each pair of mated
        FASTQ files) whose paths are given as strings. """
        path1_strs = None
        for fq_key, path_strs in fastq_args.items():
            try:
                # Check if fq_key is a key for a directory or file.
                dir_type = cls.KeySampleDirType[
                    cls.KeySampleDirName(fq_key).name].value
            except ValueError:
                # fq_key is not a directory key: assume a file key.
                if (fq_key == cls.KeyName.MATE1.value
                        or fq_key == cls.KeyName.MATE2.value):
                    # fq_key is for files of mate 1 or mate 2 reads.
                    if path1_strs is None:
                        # Because both keys for mate 1 and mate 2 must
                        # be read together to generate the FASTQ units,
                        # confirm that both keys were given.
                        # To prevent yielding each of these FASTQ units
                        # twice (once when fq_key is mate 1 and again
                        # when fq_key is mate 2), this block runs only
                        # if path1_strs is None (which is no longer true
                        # after the first time this block runs).
                        try:
                            path1_strs = fastq_args[cls.KeyName.MATE1.value]
                            path2_strs = fastq_args[cls.KeyName.MATE2.value]
                        except KeyError:
                            raise KeyError(
                                f"Must give both '{cls.KeyName.MATE1.value}' "
                                f"and '{cls.KeyName.MATE2.value}', or neither")
                        #
                        yield from cls._get_sample_pairs(phred_enc=phred_enc,
                                                         path1_strs=path1_strs,
                                                         path2_strs=path2_strs)
                else:
                    # fq_key is for files of single-end or interleaved
                    # paired-end reads. Yield every file from the key.
                    yield from cls._get_sample_files(phred_enc=phred_enc,
                                                     path_strs=path_strs,
                                                     key=fq_key)
            else:
                # The FASTQ key is a key for a directory.
                # For each path string, parse the directory path.
                for dir_path in map(dir_type.parse, path_strs):
                    # Yield each FastqUnit depending on the FASTQ key.
                    if fq_key == cls.KeySampleDirName.MATE12.value:
                        # The directory contains paired reads with mate
                        # 1 and mate 2 reads in separate FASTQ files.
                        yield from cls._get_demult_pairs(phred_enc=phred_enc,
                                                         dir_path=dir_path)
                    else:
                        # The directory contains FASTQ files that may be
                        # processed separately (single-end/interleaved).
                        yield from cls._get_demult_files(phred_enc=phred_enc,
                                                         dir_path=dir_path,
                                                         key=fq_key)

    @classmethod
    def from_strs(cls, /, *,
                  phred_enc: int,
                  no_dup_samples: bool = True,
                  **fastq_args: tuple[str]):
        """
        Yield a FastqUnit for each FASTQ file (or each pair of mate 1
        and mate 2 FASTQ files) whose paths are given as strings.

        Parameters
        ----------
        phred_enc: int
            ASCII offset for encoding Phred scores
        no_dup_samples: bool (default: True)
            If a sample name occurs more than once among the given FASTQ
            files, then log a warning and yield only the first FASTQ
            with that sample name.
        fastq_args: tuple[str]
            FASTQ files, given as tuples of file path string:
            - fastqs: FASTQ files of single-end reads
            - fastqi: FASTQ files of interleaved paired-end reads
            - fastq1: FASTQ files of mate 1 paired-end reads; must
                      correspond 1-for-1 (in order) with fastq2
            - fastq2: FASTQ files of mate 2 paired-end reads; must
                      correspond 1-for-1 (in order) with fastq1
            - fastqs_dir: Directory of FASTQ files of single-end reads
            - fastqi_dir: Directory of FASTQ files of interleaved
                          paired-end reads
            - fastq12_dir: Directory of FASTQ files of separate mate 1
                           and mate 2 paired-end reads; for every FASTQ
                           file of mate 1 reads, there must be a FASTQ
                           file of mate 2 reads with the same sample
                           name, and vice versa.

        Yield
        -----
        FastqUnit
            FastqUnit representing the FASTQ or pair of FASTQ files.
            The order is determined primarily by the order of keyword
            arguments; within each keyword argument, by the order of
            file or directory paths; and for directories, by the order
            in which ```pathlib.Path.iterdir``` returns file paths.
        """
        samples = set()
        for fq_unit in cls._from_strs(phred_enc=phred_enc, **fastq_args):
            if no_dup_samples:
                if (sample := fq_unit.sample) in samples:
                    logging.warning(f"Skipping duplicate sample: '{sample}'")
                    continue
                samples.add(sample)
            yield fq_unit

    def __str__(self):
        return " and ".join(self.str_paths)


class ReadsFileBase(object):
    _module: path.Module | None = None
    _step: str | None = None
    _ext: str | None = None

    def __init__(self, /, *,
                 input_path: (FastqUnit |
                              path.AbstractRefsetAlignmentFilePath |
                              path.AbstractOneRefAlignmentFilePath),
                 top_dir: str,
                 n_procs: int,
                 save_temp: bool,
                 resume: bool):
        self._input = input_path
        self._top_dir = path.TopDirPath.parse(top_dir)
        self._n_procs = n_procs
        self._save_temp = save_temp
        self._resume = resume
        self._output_files: list[path.BasePath] = list()
        self._index_files: list[path.BasePath] = list()

    @property
    def output_fields(self):
        return {**self._input.dict(),
                **dict(top=self._top_dir.top,
                       module=self._module,
                       step=self._step,
                       ext=self._ext)}

    @property
    def output(self):
        return self._input.edit(**self.output_fields)

    def _setup(self):
        self.output.parent.path.mkdir(parents=True, exist_ok=True)

    def _run(self, **kwargs):
        """ Run this step. Each class implements a unique function. """
        logging.error("The base class does not implement a run method")

    def _clean_index_files(self):
        while self._index_files:
            self._index_files.pop().path.unlink(missing_ok=True)

    def _needs_to_run(self):
        return not self._resume or not self.output.path.is_file()

    def run(self, **kwargs):
        """ Wrapper for internal run method that first checks if the
        step actually needs to be run because its output files might
        already exist. """
        if self._needs_to_run():
            # If the output of this step does not exist, or it does
            # (because, e.g. the alignment pipeline crashed after at
            # least one step had written a temporary output file), but
            # the pipeline is not allowed to resume from the leftover
            # temporary files, then run this step.
            self._setup()
            self._run(**kwargs)
            if not self._save_temp:
                self._clean_index_files()
        # Return the output of this step so that it can be fed into the
        # next step.
        return self.output

    def clean(self):
        if self._step is not None and not self._save_temp:
            while self._output_files:
                self._output_files.pop().path.unlink(missing_ok=True)


class FastqBase(ReadsFileBase):
    _module = path.Module.ALIGN

    def __init__(self, /, *, fq_unit: FastqUnit, **kwargs):
        super().__init__(**kwargs, input_path=fq_unit)

    @property
    def sample(self):
        return self._input.sample

    @staticmethod
    def _qc(fastq_paths: list[path.BasePath], extract: bool):
        cmd = [FASTQC_CMD]
        if extract:
            cmd.append("--extract")
        cmd.extend(fastq_paths)
        run_cmd(cmd)

    def qc(self, extract: bool):
        self._qc(list(self._input.inputs.values()), extract)


class FastqTrimmer(FastqBase):
    _step = path.Step.ALIGN_TRIM
    _cutadapt_output_flags = "-o", "-p"

    @cached_property
    def output(self) -> FastqUnit:
        return self._input.edit(**self.output_fields)

    @property
    def _cutadapt_output_args(self):
        """ Return arguments that specify Cutadapt output files. """
        return tuple(itertools.chain(*zip(self._cutadapt_output_flags,
                                          self.output.paths,
                                          strict=False)))

    @docdef.auto()
    def _cutadapt(self, /, *,
                  cut_q1: int,
                  cut_q2: int,
                  cut_g1: str,
                  cut_a1: str,
                  cut_g2: str,
                  cut_a2: str,
                  cut_o: int,
                  cut_e: float,
                  cut_indels: bool,
                  cut_nextseq: bool,
                  cut_discard_trimmed: bool,
                  cut_discard_untrimmed: bool,
                  cut_m: int):
        """ Trim adapters and low-quality bases with Cutadapt. """
        cmd = [CUTADAPT_CMD]
        cmd.extend(["--cores", self._n_procs])
        if cut_nextseq:
            if cut_q1 > 0:
                cmd.extend(["--nextseq-trim", cut_q1])
        else:
            if cut_q1 > 0:
                cmd.extend(["-q", cut_q1])
            if cut_q2 > 0:
                cmd.extend(["-Q", cut_q2])
        adapters = {"g": cut_g1, "a": cut_a1,
                    "G": cut_g2, "A": cut_a2}
        for arg, adapter in adapters.items():
            if adapter and (self._input.paired or arg.islower()):
                for adapt in adapter:
                    cmd.extend([f"-{arg}", adapt])
        cmd.extend(["-O", cut_o])
        cmd.extend(["-e", cut_e])
        cmd.extend(["-m", cut_m])
        if not cut_indels:
            cmd.append("--no-indels")
        if cut_discard_trimmed:
            cmd.append("--discard-trimmed")
        if cut_discard_untrimmed:
            cmd.append("--discard-untrimmed")
        if self._input.interleaved:
            cmd.append("--interleaved")
        cmd.extend(["--report", "minimal"])
        cmd.extend(self._cutadapt_output_args)
        cmd.extend(self._input.cutadapt_input_args)
        run_cmd(cmd)
        self._output_files.extend(self.output.inputs.values())

    def _needs_to_run(self):
        return not self._resume or not all(output.is_file()
                                           for output in self.output.paths)

    def _run(self, **kwargs):
        self._cutadapt(**kwargs)


class FastqAligner(FastqBase):
    _step = path.Step.ALIGN_ALIGN
    _ext = path.SAM_EXT

    def __init__(self, /, *,
                 fasta: (path.AbstractRefsetSeqFilePath |
                         path.OneRefSeqStepFilePath),
                 **kwargs):
        super().__init__(**kwargs)
        if not self._input.compatible_fasta(fasta):
            raise TypeError(f"FASTA {fasta} is not compatible with "
                            f"FASTQ file(s) {self._input}")
        self._fasta = fasta

    @cached_property
    def output(self):
        return path.create(**{**self._input.dict(),
                              **self._fasta.dict(),
                              **self.output_fields})

    @cached_property
    def _bowtie2_index_files(self):
        return [self._fasta.edit(ext=ext)
                for ext in path.BOWTIE2_INDEX_EXTS]

    @cached_property
    def _fasta_prefix(self):
        prefix = str(self._fasta.path.with_suffix(""))
        if bad_idxs := [index for index in map(str, self._bowtie2_index_files)
                        if not index.startswith(prefix)]:
            logging.critical(f"Bowtie2 index files {bad_idxs} do not start "
                             f"with FASTA prefix '{prefix}'")
        return prefix

    @property
    def _missing_bowtie2_index_files(self):
        return [index for index in self._bowtie2_index_files
                if not index.path.is_file()]

    def _bowtie2_build(self):
        """ Build an index of a reference using Bowtie 2. """
        cmd = [BOWTIE2_BUILD_CMD, self._fasta, self._fasta_prefix]
        run_cmd(cmd)
        self._index_files.extend(self._bowtie2_index_files)

    @docdef.auto()
    def _bowtie2(self, /, *,
                 bt2_local: bool,
                 bt2_discordant: bool,
                 bt2_mixed: bool,
                 bt2_dovetail: bool,
                 bt2_contain: bool,
                 bt2_unal: bool,
                 bt2_score_min: str,
                 bt2_i: int,
                 bt2_x: int,
                 bt2_gbar: int,
                 bt2_l: int,
                 bt2_s: str,
                 bt2_d: int,
                 bt2_r: int,
                 bt2_dpad: int,
                 bt2_orient: str):
        """ Align reads to the reference with Bowtie 2. """
        if missing := self._missing_bowtie2_index_files:
            logging.critical("Bowtie2 index files do not exist:\n\n"
                             + "\n".join(map(str, missing)))
            return
        cmd = [BOWTIE2_CMD]
        # Resources
        cmd.extend(["--threads", self._n_procs])
        # Alignment
        if bt2_local:
            cmd.append("--local")
        cmd.extend(["--gbar", bt2_gbar])
        cmd.extend(["--dpad", bt2_dpad])
        cmd.extend(["-L", bt2_l])
        cmd.extend(["-i", bt2_s])
        cmd.extend(["-D", bt2_d])
        cmd.extend(["-R", bt2_r])
        # Scoring
        cmd.append(self._input.phred_arg)
        cmd.append("--ignore-quals")
        cmd.extend(["--ma", MATCH_BONUS])
        cmd.extend(["--mp", MISMATCH_PENALTY])
        cmd.extend(["--np", N_PENALTY])
        cmd.extend(["--rfg", REF_GAP_PENALTY])
        cmd.extend(["--rdg", READ_GAP_PENALTY])
        # Filtering
        if not bt2_unal:
            cmd.append("--no-unal")
        cmd.extend(["--score-min", bt2_score_min])
        cmd.extend(["-I", bt2_i])
        cmd.extend(["-X", bt2_x])
        # Mate pair orientation
        orientations = tuple(op.value for op in cli.MateOrientationOption)
        if bt2_orient in orientations:
            cmd.append(f"--{bt2_orient}")
        else:
            cmd.append(f"--{orientations[0]}")
            logging.warning(f"Invalid mate orientation: '{bt2_orient}'; "
                            f"defaulting to '{orientations[0]}'")
        if not bt2_discordant:
            cmd.append("--no-discordant")
        if not bt2_contain:
            cmd.append("--no-contain")
        if bt2_dovetail:
            cmd.append("--dovetail")
        if not bt2_mixed:
            cmd.append("--no-mixed")
        # Formatting
        cmd.append("--xeq")
        # Input and output files
        cmd.extend(["-S", self.output.path])
        cmd.extend(["-x", self._fasta_prefix])
        cmd.extend(self._input.bowtie2_inputs)
        # Run alignment.
        run_cmd(cmd)
        self._output_files.append(self.output)

    def _run(self, **kwargs):
        if self._missing_bowtie2_index_files:
            self._bowtie2_build()
        self._bowtie2(**kwargs)


class XamBase(ReadsFileBase):
    def view(self, output: (path.AbstractRefsetAlignmentFilePath |
                            path.AbstractOneRefAlignmentFilePath |
                            path.AbstractRegionAlignmentFilePath |
                            path.BasePath)):
        cmd = [SAMTOOLS_CMD, "view"]
        if self._ext == path.BAM_EXT:
            cmd.append("-b")
        cmd.extend(("-o", output, self._input))
        if isinstance(output, path.AbstractRefsetAlignmentFilePath):
            # View all reads. No index is required.
            pass
        elif isinstance(output, path.AbstractOneRefAlignmentFilePath):
            # View only reads that aligned to the reference.
            cmd.append(output.ref)
            # This restriction requires an index.
            self._build_index(self._input)
        elif isinstance(output, (path.RegionAlignmentInFilePath |
                                 path.RegionAlignmentStepFilePath |
                                 path.RegionAlignmentOutFilePath)):
            # View only reads that aligned to a region of the reference.
            cmd.append(f"{output.ref}:{output.end5}-{output.end3}")
            # This restriction requires an index.
            self._build_index(self._input)
        else:
            raise TypeError(output)
        run_cmd(cmd)
        if output.path.is_file():
            self._output_files.append(output)
        else:
            logging.critical(f"Failed to view file: {self._input} -> {output}")

    @staticmethod
    def _get_index_path(bam: path.BasePath):
        return bam.edit(ext=path.BAI_EXT)

    def _build_index(self, bam: path.BasePath):
        bam_index = self._get_index_path(bam)
        if not self._resume or not bam_index.path.is_file():
            index_bam_file(bam)
            self._index_files.append(bam_index)


class SamRemoveEqualMappers(XamBase):
    _module = path.Module.ALIGN
    _step = path.Step.ALIGN_REMEQ
    _ext = path.SAM_EXT

    pattern_a = re.compile(SAM_ALIGN_SCORE + rb"(\d+)")
    pattern_x = re.compile(SAM_EXTRA_SCORE + rb"(\d+)")

    _MIN_SAM_FIELDS = 11
    _MAX_SAM_FLAG = 4095  # 2^12 - 1

    @staticmethod
    def _get_score(line: bytes, ptn: re.Pattern[bytes]):
        return (float(match.groups()[0])
                if (match := ptn.search(line)) else None)

    @classmethod
    def _is_best_alignment(cls, line: bytes):
        return ((score_x := cls._get_score(line, cls.pattern_x)) is None
                or score_x < cls._get_score(line, cls.pattern_a))

    @classmethod
    def _read_is_paired(cls, line: bytes):
        info = line.split()
        if len(info) < cls._MIN_SAM_FIELDS:
            raise ValueError(f"Invalid SAM line:\n{line.decode()}")
        flag = int(info[1])
        if 0 <= flag <= cls._MAX_SAM_FLAG:
            return bool(flag % 2)
        raise ValueError(f"Invalid SAM flag: {flag}")

    def _iter_paired(self, sam: BinaryIO, line: bytes):
        """ For each pair of reads, yield the pair of alignments for
        which both the forward alignment and the reverse alignment in
        the pair scored best among all alignments for the forward and
        reverse reads, respectively. Exclude pairs for which the forward
        and/or reverse read aligned equally well to multiple locations,
        or for which the best alignments for the forward and reverse
        reads individually are not part of the same alignment pair. """
        total_lines = 0
        yield_lines = 0
        while line:
            total_lines += 1
            line2 = sam.readline()
            total_lines += bool(line2)
            if self._is_best_alignment(line) and self._is_best_alignment(line2):
                yield_lines += 1
                yield line
                yield_lines += 1
                yield line2
            line = sam.readline()
        if total_lines % 2:
            logging.error(f"SAM file {self._input} was paired but had an odd "
                          f"number of lines ({total_lines})")

    def _iter_single(self, sam: BinaryIO, line: bytes):
        """ For each read, yield the best-scoring alignment, excluding
        reads that aligned equally well to multiple locations. """
        while line:
            if self._is_best_alignment(line):
                yield line
            line = sam.readline()

    def _remove_equal_mappers(self, buffer_length):
        with (open(self._input.path, "rb") as sami,
              open(self.output.path, "wb") as samo):
            # Copy the header from the input to the output SAM file.
            while (line := sami.readline()).startswith(SAM_HEADER):
                samo.write(line)
            if line:
                if self._read_is_paired(line):
                    lines = self._iter_paired(sami, line)
                else:
                    lines = self._iter_single(sami, line)
                while text := b"".join(itertools.islice(lines, buffer_length)):
                    samo.write(text)

    def _run(self, rem_buffer: int = cli.opt_rem_buffer.default):
        logging.info("\nRemoving Reads Mapping Equally to Multiple Locations"
                     f" in {self._input.path}\n")
        self._remove_equal_mappers(rem_buffer)
        self._output_files.append(self.output)


class XamSorter(XamBase):
    def _sort(self, name: bool):
        cmd = [SAMTOOLS_CMD, "sort"]
        if name:
            cmd.append("-n")
        cmd.extend(["-o", self.output.path, self._input.path])
        run_cmd(cmd)
        self._output_files.append(self.output)

    def _run(self, name: bool = False):
        logging.info(f"\nSorting {self._input.path} by Reference and Coordinate\n")
        self._sort(name)


class BamAlignSorter(XamSorter):
    _module = path.Module.ALIGN
    _step = path.Step.ALIGN_SORT
    _ext = path.BAM_EXT


class SamVectorSorter(XamSorter):
    _module = path.Module.VECTOR
    _step = path.Step.VECTOR_SORT
    _ext = path.SAM_EXT


class BamSplitter(XamBase):
    _module = path.Module.ALIGN
    _step = path.Step.ALIGN_SPLIT
    _ext = path.BAM_EXT

    def __init__(self, /, *,
                 fasta: path.RefsetSeqInFilePath | path.OneRefSeqStepFilePath,
                 **kwargs):
        super().__init__(**kwargs)
        self.fasta = fasta

    @cached_property
    def refs(self):
        return tuple(ref for ref, _ in FastaParser(self.fasta.path).parse())

    @cached_property
    def output(self):
        output = [self._input.edit(**{**self.output_fields,
                                      **dict(refset=None, ref=ref)})
                  for ref in self.refs]
        return output

    def _split(self):
        for output in self.output:
            self.view(output)

    def _needs_to_run(self):
        return not self._resume or not all(output.path.is_file()
                                           for output in self.output)

    def _setup(self):
        for output in self.output:
            output.parent.path.mkdir(parents=True, exist_ok=True)

    def _run(self):
        logging.info(f"\nSplitting {self._input} by reference\n")
        self._build_index(self._input)
        self._split()


class BamOutputter(XamBase):
    _module = path.Module.ALIGN
    _ext = path.BAM_EXT

    def _run(self):
        logging.info(f"\nOutputting {self._input} to {self.output}\n")
        self.view(self.output)


class BamVectorSelector(XamBase):
    _module = path.Module.VECTOR
    _step = path.Step.VECTOR_SELECT
    _ext = path.BAM_EXT

    def __init__(self, /, *, ref: str, end5: int, end3: int, **kwargs):
        super().__init__(**kwargs)
        self.ref = ref
        self.end5 = end5
        self.end3 = end3

    @staticmethod
    def ref_coords(ref: str, end5: int, end3: int):
        return f"{ref}:{end5}-{end3}"

    @property
    def fields(self):
        return dict(**super().output_fields,
                    end5=self.end5,
                    end3=self.end3)

    def _run(self):
        cmd = [SAMTOOLS_CMD, "view", "-h", "-o", self.output,
               self._input, self.ref_coords(self.ref, self.end5, self.end3)]
        run_cmd(cmd)


'''
primer1 = "CAGCACTCAGAGCTAATACGACTCACTATA"
primer1rc = "TATAGTGAGTCGTATTAGCTCTGAGTGCTG"
primer2 = "TGAAGAGCTGGAACGCTTCACTGA"
primer2rc = "TCAGTGAAGCGTTCCAGCTCTTCA"
adapters5 = (primer1, primer2rc)
adapters3 = (primer2, primer1rc)
'''



import os, sys
import pandas as pd

from .library_samples import get_samples_info, get_library_info
#sys.path.append(os.path.dirname(__file__))
from .mutation_count import generate_mut_profile_from_bit_vector
from ..util.files_sanity import check_library, check_samples
from ..util.rnastructure import RNAstructure
from ..util.seq import parse_fasta
from ..util.dump import *
import logging
from ..util import docdef
from ..util import path

@docdef.auto()
def run( 
        fasta: str, 
        bv_files: list,
        *,
        library: str, 
        samples: str, 
        sample: str, 
        clustering_file: str, 
        out_dir: str, 
        rnastructure_path: str, 
        verbose: bool):
    """Run the aggregate module.

    Reads in the bit vector files and aggregates them into a single file named [output]/output/aggregate/[name].csv.
    Adds on the library information, the sample information, and RNAstructure predictions.

    Parameters from args:
    ---------------------

    bv_files: str
        Path to the bit vector files folders.
    library: str
        Path to a csv file with the library information.
    samples: str
        Path to a csv file with the sample information.
    sample: str
        Name to identify the row in samples.csv. Also the name for the output file. Default is the containing folder name.
    clustering_file: str
        Path to the clustering.json file.
    out_dir: str
        Path to the output folder (the sample).
    fasta: str
        Path to the fasta file.
    verbose: bool
        Verbose output.

    Returns:
    --------
    1 if successful, 0 otherwise.

    """

    # Extract the arguments
    library = check_library(pd.read_csv(library), fasta, out_dir) if library is not None else None
    library['section_boundaries'] = library.apply(lambda x: str(x['section_start']) + '-' + str(x['section_end']), axis=1)
    df_samples = check_samples(pd.read_csv(samples)) if samples != '' else None
    fasta = pd.DataFrame({k :v.decode("utf-8")  for k,v in parse_fasta(fasta)}, index=[0]).T.reset_index().rename(columns={"index":"reference", 0:"sequence"})
    
    # Find a name for the sample
    if sample is None:
        if df_samples is not None:
            raise ValueError('If samples is specified, sample must also be specified.')
        sample = 'output'
    
    # Make folders
    os.makedirs(out_dir, exist_ok=True)
    os.makedirs(os.path.join(out_dir, 'temp'), exist_ok=True)

    # Read in the bit mut_vectors
    if clustering_file is not None:
        with open(clustering_file, 'r') as f:
            clustering_file = json.load(f)    
    
    mut_profiles = {}
    print('Reading in bit mut_vectors from {}...'.format(bv_files))
    
    bv_files = [b.replace('_report.json', '') for b in bv_files if b.endswith('_report.json')]
    
    for bv in bv_files:
        
        reference, section = bv.split('/')[-2], bv.split('/')[-1]
        
        assert len(library[(library['reference'] == reference)&(library['section_boundaries'] == section)]) < 2, 'Library information not unique for reference {} section {}'.format(reference, section)
        assert len(library[(library['reference'] == reference)&(library['section_boundaries'] == section)]) > 0, 'Library information not existing for reference {} section {}'.format(reference, section)
        section = library[(library['reference'] == reference)&(library['section_boundaries'] == section)]['section'].values[0]
        
        if not len(os.listdir(bv)) > 0:
            logging.warning('No bit vectors found for reference {}'.format(reference))
            continue
        
        if reference not in mut_profiles:
            mut_profiles[reference] = {'sequence': fasta[fasta['reference'] == reference]['sequence'].values[0]}
        # Add the library information
        mut_profiles[reference] = {**get_library_info(library, reference, verbose=verbose), **mut_profiles[reference]}
        mut_profiles[reference].pop('section_boundaries')
        
        assert library[(library['reference'] == reference)&(library['section'] == section)].shape[0] == 1, 'Library information not found for reference {} section {}'.format(reference, section)
        mut_profiles[reference][section] = {}
        mut_profiles[reference][section]['section_start'] = library[(library['reference'] == reference)&(library['section'] == section)]['section_start'].values[0]
        mut_profiles[reference][section]['section_end'] = library[(library['reference'] == reference)&(library['section'] == section)]['section_end'].values[0]
        mut_profiles[reference][section]['pop_avg'] = generate_mut_profile_from_bit_vector(bv, clustering_file=clustering_file, verbose=verbose)
        
        # TODO
        mut_profiles[reference][section]['pop_avg'].pop('sequence')
        mut_profiles[reference][section]['sequence'] = mut_profiles[reference]['sequence'][mut_profiles[reference][section]['section_start']-1:mut_profiles[reference][section]['section_end']]# mut_profiles[reference][section]['pop_avg'].pop('sequence')
        # assert mut_profiles[reference]['sequence'][mut_profiles[reference][section]['section_start']-1:mut_profiles[reference][section]['section_end']] == mut_profiles[reference][section]['sequence'], 'Sequence mismatch for reference {} section {}: {} vs {}'.format(reference, section, mut_profiles[reference]['sequence'][mut_profiles[reference][section]['section_start']-1:mut_profiles[reference][section]['section_end']], mut_profiles[reference][section]['sequence'])
        
        
        for col in ['num_aligned']:
            mut_profiles[reference][col] = mut_profiles[reference][section]['pop_avg'].pop(col)

    print('Done.')
    if df_samples is not None:
        # Add the sample information
        print('Adding sample information...')
        mut_profiles = {**mut_profiles, **get_samples_info(df_samples, sample, verbose=verbose)}
        print('Done.')
        
    
    print('Computing confidence intervals and RNAstructure predictions...')
    rna = RNAstructure(rnastructure_path=rnastructure_path)
    for reference in mut_profiles:
        if type(mut_profiles[reference]) is not dict:
            continue

        for section in mut_profiles[reference]:
            if type(mut_profiles[reference][section]) is not dict:
                continue
            # Add RNAstructure predictions

            mh = rna.run(mut_profiles[reference][section]['sequence'])
            mut_profiles[reference][section] = {**mut_profiles[reference][section], **mh}

    print('Done.')
    # Write the output
    print('Cast dictionary, size:', sys.getsizeof(mut_profiles))
    out = cast_dict(mut_profiles)
    print('Done.')
    print('Sort dictionary, size:', sys.getsizeof(mut_profiles))
    out = sort_dict(out)
    print('Done.')
    print('Dump the json, size', sys.getsizeof(json.dumps(out, cls=NpEncoder)))
    with open(os.path.join(out_dir, sample + '.json'), 'w') as f:
        f.write(json.dumps(out, cls=NpEncoder, indent=2))
    print('Done.')
    print('Done aggregating the data.')
    return 1



from setuptools import setup, find_packages
#from dreem import __version__
import sys, os

requirements_path = os.path.join(os.getcwd(), 'requirements.txt')
requirements_path = '/Users/ymdt/src/dreem/requirements.txt'
print("Python version: {}".format(sys.version))
requirements = []
with open(requirements_path, 'r') as fh:
    for line in fh:
        requirements.append(line.strip())


#PYTHON_VERSION = (3,10)

#if sys.version_info < PYTHON_VERSION:
#    sys.exit(f"Python >= {PYTHON_VERSION[0]}.{PYTHON_VERSION[1]} required.")

readme = open('README.md').read()

setup(
    version='0.1.0',
  long_description=readme,
   packages=find_packages(),
   package_dir={'dreem': 'dreem'},
   py_modules=[
        'dreem',
        'dreem/demultiplex',
        'dreem/align',
        'dreem/vector',
        'dreem/cluster',
        'dreem/aggregate',
        'dreem/draw',
        'dreem/test',
        'dreem/util',
   ],
   include_package_data=True,
   install_requires=requirements,
   entry_points = {'console_scripts' : ['dreem = dreem.main : cli']}
)
